#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <EzGraph.h>
#include "3dshoot.h"
#include "motion.h"
#include "2d.h"
#include "color.h"
#include "file.h"
#include "getarg.h"
#include "uty.h"
#include "rnd.h"

myship_t myship = {{300, FLIGHT_HEIGHT, 850}};
bullet_t mybullet[MYBULLET_MAX];
int enemyship_num;
enemyship_t enemyship[ENEMYSHIP_MAX];
bullet_t enemybullet[ENEMYBULLET_MAX];
breath_t breath[5][BREATH_LEN];
laser_t lasers[LASER_MAX_NUM];
enemyship_t *dragon;

static unsigned char major_version = MAJOR_VERSION;
static unsigned char minor_version = MINOR_VERSION;

unsigned int dir_keymap = 0;
int dir_table[] = {
	0, 4, 8, 7,
	6, 0, 9, 0,
	2, 1, 0, 0,
	3, 0, 0, 0
};

point2d_t move_vec[] = {
    { 0,  0},  // 0: ダミー
    {-1, -1},  // 1
    { 0, -1},  // 2
    { 1, -1},  // 3
    {-1,  0},  // 4
    { 0,  0},  // 5
    { 1,  0},  // 6
    {-1,  1},  // 7
    { 0,  1},  // 8
    { 1,  1},  // 9
};
int dir, dir_prev, is_fire = 0;

//int is_autofire = 1;
int is_autofire = 0;
int is_pause = 0;
int is_mortal = 1;
int is_cheat = 0;
unsigned int score = 0;
//int hi_score = 0;
gamestate_t gamestate = gs_title;
// グリッドアニメーション・点滅用カウンタ(3 -> 2 -> 1 -> 0 -> 3 -> ,,,)
int grid_counter= 0;
// ゲーム状態の遷移タイミング制御
int gamestate_counter = 0;
int boss_life, boss_life_max;
int dragon_head_num;
int stage;
int immortal_counter = 0;
cmd_t cmd = cm_start;
achievements_t achievements = {{0}};
int game_speed;

int info_lt_x, info_lt_y;
int sensor_lt_x, sensor_lt_y;


// 表示名と得点
enemydata_t enemydata[] = {
  {"SCOUT",      100},
  {"WALL",      9999},
  {"DRONE",      200},
  {"FIGHTER",    300},
  {"COPTER",     300},
  {"LANCER",     200},
  {"WALKER",     500},
  {"TURRET",     500},
  {"FORTRESS", 10000},
  {"BOMBER",    3000},
  // BYTEは一機撃破毎に点が入るので点数が低め
  {"BYTE",       500},
  {"GUNROID",  15000},
  {"WYVERN",   10000},
  {"DRAGON",   10000},
  {"HEAD",      3000},
  {"CORE",     10000},
};

group_t group_original[] = {
  // オリジナル版の面構成
  {et_scout, 10, 3},
  {et_lancer, 8, 2},
  {et_copter, 4, 3},
  {et_wall, 10, 6},
  {et_lancer, 8, 4},
  {et_scout, 4, 3},
  {et_copter, 2, 3},
  {et_walker, 3, 3},
  {et_lancer, 4, 4},
  {et_scout, 8, 4},
  {et_copter, 4, 4},
  {et_wall, 2, 4},
  {et_walker, 1, 4},
  {et_scout, 3, 4},
  {et_walker, 1, 4},
  {et_wall, 2, 4},
  {et_walker, 1, 4},
  {et_copter, 2, 6},
  {et_wall, 2, 6},
  {et_lancer, 2, 6},
  {et_wall, 2, 6},
  {et_walker, 1, 6},
  {et_copter, 2, 6},
  {et_scout, 8, 8},
  {et_lancer, 8, 6},
  {et_copter, 8, 6},
  // fortress直前のグループは上限1にしないと、全部倒す前にfortressが出現する
  {et_wall, 1, 1},
  {et_fortress, 1, 1},
};
group_t group_extended[] = {
  {et_scout, 10, 3},
  {et_drone, 8, 2},
  {et_copter, 4, 3},
  {et_scout, 6, 4},
  {et_fighter, 5, 2},
  {et_wall, 2, 2},
  {et_copter, 2, 3},
  {et_scout, 6, 4},
  {et_turret, 2, 4},
  {et_drone, 4, 4},
  {et_lancer, 4, 4},
  // WALL地帯
  {et_scout, 4, 8},
  {et_wall,  4, 8},
  {et_drone, 8, 8},
  {et_wall,  4, 8},
  // 前半戦ラストにWALKERがちらっと出る
  {et_walker, 1, 3},
  // 中ボスは雑魚が残っていても出現させる
  {et_scout,  4, 3},
  {et_midboss, 1, 1},
  // TURRET地帯
  {et_copter,  2, 6},
  {et_turret,  4, 6},
  {et_lancer,  2, 6},
  {et_turret,  4, 6},
  {et_copter,  2, 6},
  {et_turret,  4, 6},
  {et_lancer,  2, 6},
  // WALKER多数含むラス前の猛攻
  {et_scout,   4, 6},
  {et_walker,  2, 6},
  {et_fighter, 4, 6},
  {et_lancer,  4, 4},
  {et_walker,  2, 6},
  {et_fighter, 4, 6},
  {et_turret,  2, 6},
  {et_walker,  4, 6},
  {et_copter,  2, 6},
  {et_fighter, 4, 6},
  {et_turret,  2, 6},
  {et_walker,  4, 6},
  {et_copter,  4, 6},
  {et_walker,  4, 6},
  {et_turret,  2, 6},
  {et_lancer,  8, 4},
  // ラス前のWALL地帯
  {et_wall,    7, 6},
  // ボス直前のグループは上限1にしないと、全部倒す前にボスが出現する
  {et_wall,  1, 1},
  {et_boss,1, 1},
};
group_t group_boss[] = {
    {et_midboss,1, 1},
    {et_boss,1, 1},
    //    {et_scout,  4, 4}, // !!!
    //    {et_dragon, 1, 1}, // !!!
    //    {et_boss,1, 1},
    //    {et_dragon, 1, 1}, // !!!
    //    {et_core, 1, 1}, // !!!
    /*
    {et_byte, 1, 1},
    {et_gunroid, 1, 1},
    {et_dragon, 1, 1},
    {et_bomber, 1, 1},
    {et_fortress, 1, 1},
    {et_byte, 1, 1},
    {et_gunroid, 1, 1},
    {et_wyvern, 1, 1},
    */
};

// 各モードのグループ系列とグループ数
group_t *mode_group[] = {group_original, group_extended, group_boss};
int mode_group_num[] = {
    (sizeof(group_original) / sizeof(group_t)),
    (sizeof(group_extended) / sizeof(group_t)),
    (sizeof(group_boss) / sizeof(group_t)),
};
// プレイ中のモードのグループ系列とグループ数
group_t *group;
int group_num;

int group_no;
group_t current_group;
// 現在生存中のcurrent_groupの敵の数
int current_group_alive_enemy_num;

int midboss_index, boss_index;
int extended_midboss_index, extended_boss_index;
int boss_midboss_index, boss_boss_index;


char *achievements_str[] = {
    "MODE",
    "TOP SCORE",
    "TOP STAGE",
    "TOTAL PLAY",
    "TOTAL WIN-LOSS",
    "  TOTAL",
    "  SCOUT",
    "  WALL",
    "  DRONE",
    "  FIGHTER",
    "  COPTER",
    "  LANCER",
    "  WALKER",
    "  TURRETS",
    "  FORTRESS",
    "  BOMBER",
    "  BYTE",
    "  GUNROID",
    "  WYVERN",
    "  DRAGON",
    "  DRAGON_HEAD",   // 実際は表示しない
    "  CORE",
};
#define ACHIEVEMENTS_NUM (sizeof(achievements_str) / sizeof(char*))

char *config_str[] = {
    "STAGE",
    "MY SHIP",
    "AUTO FIRE",
    "HITBOX",
    "COLOR",
    "VIEW",
};

#define CONFIG_NUM (sizeof(config_str) / sizeof(char*))

char *options_str[CONFIG_NUM+1][9] = {
    {"ORIGINAL", "EXTENDED", "BOSS", "", "", "", ""},
    {"1", "2", "3", "4", "5", "6", "7", "8", "9"},
    {"OFF", "ON"},
    {"OFF", "ON"},
    {"ORIG", "IBM", "MAC", "MZ", "INV", "", "", ""},
    {"BIRD", "TOP", "ISOMETRIC"},
    {"[END]", "", "", "", "", "", ""},
};
int options_num[] = {
    3,
    9,
    2,
    2,
    5,
    3,
    1
};

config_item_t config_item = ci_stage;
config_t config = {
    {
        0,                   // 0:ORIGINAL, 1:EXTENDED
        MYSHIP_LEFT_DEFAULT, // MY SHIP
        AUTOFIRE_DEFAULT,    // AUTO FIRE
        SHOW_HITBOX_DEFAULT, // SHOW HITBOX
        cs_original,         // COLOR
        VIEW_MODE_DEFAULT,   // VIEW
        0                    // END
    }
};

char boss_pass[] = {'c', 111, 0x66, 0146, 0x63, 121, 0};
int boss_ix = 0;
//char *exe_dir;


int main(int argc, char *argv[]){

    init(argc, argv);
    EzEventLoop();
    quit();
    return 0;
}

// p1とp2の距離がrange以下なら非0
// 距離は矩形領域で判定
// とりあえずY軸は考えずXZ平面だけで判定
int is_hit(point3d_t p1, point3d_t p2, int xrange, int zrange){
    if (abs(p1.x - p2.x) <= xrange && abs(p1.z - p2.z) <= zrange){
        printf("hit\n");
        return 1;
    }
    else {
        return 0;
    }
}

void title_timerh(void){
    //  static int blink = 0;
    char buf[64];
    draw_ground();
    //    SetColor(5);
    set_color(ct_title);
    EzDrawStringB(280, 120,  "3D SHOOTING");
    sprintf(buf, "Ver %d.%02d", major_version, minor_version);
    EzDrawStringB(280+16*3/4, 144, buf);
    
    //    SetColor(4);
    set_color(ct_menu);
    EzDrawStringB(320-16*5/4,  320, "START");
    EzDrawStringB(320-16*12/4, 352, "ACHIEVEMENTS");
    EzDrawStringB(320-16*13/4, 384, "CONFIGURATION");
    EzDrawStringB(320-16*4/4,  416, "QUIT");
    EzDrawBoxB(320-16*14/4, cmd*32+320-20, 320+16*14/4, cmd*32+320+4);

    EzShowBuffer();
    //    EzSetTimerHandler(title_timerh, GAME_SPEED);
}


void title_keyh(int key){
    switch (key){
    case EZ_Down:
        cmd = (cmd + 1) % 4;
        break;
    case EZ_Up:
        cmd = (cmd + 3) % 4;
        break;
    case 's':
        cmd = cm_start;
        is_cheat = 0;
        is_mortal = 1;
        init_game();
        break;
    case 'o':
        cmd = cm_start;
        is_cheat = 1;
        is_mortal = 0;
        init_game();
        break;
    case 'a':
        cmd = cm_achievements;
        init_achievements();
        break;
    case 'c':
        cmd = cm_config;
        init_config();
        break;
    case 'q':
        EzExitEventLoop();
        //        quit();
        break;
    case ' ':
    case 13:
        switch (cmd){
        case cm_start:
            is_cheat = 0;
            is_mortal = 1;
            init_game();
            break;
        case cm_achievements:
            init_achievements();
            break;
        case cm_config:
            init_config();
            break;
        case cm_quit:
            EzExitEventLoop();
        //            quit();
            break;
        default:
            break;
        }
    default:
        break;
    }
}


void achievements_timerh(void){
    draw_ground();
    draw_achievements();
    EzShowBuffer();
    //    EzSetTimerHandler(achievements_timerh, GAME_SPEED);
}


void achievements_keyh(int key){
    switch (key){
    default:
        init_title();
        break;
    }
}


void config_timerh(void){
    int x;

    draw_ground();
    draw_config();
    x = config.options[config_item] * 300 / options_num[config_item];
    //    SetColor(7);
    set_color(ct_menu_hilight);
    EzDrawBoxB(240-4+x, config_item*32+60-20,
               240-24+x+300/options_num[config_item],
               config_item*32+60+4);
    EzShowBuffer();
}


void config_keyh(int key){
    switch (key){
    case EZ_Down:
        config_item = (config_item + 1) % (CONFIG_NUM + 1);
        break;
    case EZ_Up:
        config_item = (config_item + CONFIG_NUM + 1 - 1) % (CONFIG_NUM + 1);
        break;
    case EZ_Right:
        config.options[config_item]
            = (config.options[config_item] + 1) % options_num[config_item];
        set_color_set_type(config.options[ci_colorset]);
        set_view(config.options[ci_view]);
        break;
    case EZ_Left:
        config.options[config_item]
            = (config.options[config_item] + options_num[config_item] - 1) 
            % options_num[config_item];
        set_color_set_type(config.options[ci_colorset]);
        set_view(config.options[ci_view]);
        break;
    case ' ':
    case 13:
        if (config_item < CONFIG_NUM){
            config_item = (config_item + 1) % (CONFIG_NUM + 1);
        }
        else {
            init_title();            
        }
        break;
    case 'q':
        init_title();
        break;
    default:
        break;
    }
}

void game_timerh(void){
    int i;
    
    if (!is_pause){
        // check hit
        for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
            if (enemyship[i].state == ss_alive){
                //                printf("check hit enemy%d\n", i);
                check_hit_enemyship(&enemyship[i]);
            }
        }
        if (myship.state == ss_alive){
            for (i = 0 ; i < ENEMYBULLET_MAX ; i++){
                if (enemybullet[i].state == bs_alive){
                    check_hit_enemybullet(&enemybullet[i]);
                }
            }
        }

        move_myship();

        for (i = 0 ; i < MYBULLET_MAX ; i++){
            if (mybullet[i].state == bs_alive){
                move_mybullet(&mybullet[i]);
            }
        }
        if (enemyship_num < current_group.enemy_max){
            if (Prob(15)){
                create_enemyship();
            }
        }
        for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
            if (enemyship[i].state != ss_dead){
                move_enemyship(&enemyship[i], i);
            }
        }
        for (i = 0 ; i < ENEMYBULLET_MAX ; i++){
            if (enemybullet[i].state == bs_alive){
                move_enemybullet(&enemybullet[i]);
            }
        }
        if (current_group.type == et_core){
            move_lasers();
        }
        animate_sensor();
    }

    // 最後の一機の爆発中はgame_speedが変化するので毎回タイマをかけ直す
    if (myship.state == ss_exploding && myship.left == 0){
        printf("EXPLODING %d\n", game_speed);
        EzSetTimerHandler(game_timerh, game_speed);
    }
    else if (myship.state == ss_dead){
        gamestate_counter ++;
        if (gamestate_counter > 40){
            init_title();
            return;
        }
        /*        
        if (gamestate_counter < 17){
            ix = gamestate_counter;
        }
        else {
            ix = 17;
        }
        strncpy(buf, "G A M E  O V E R", ix);
        buf[ix] = '\0';
        //        SetColor(2);
        set_color(ct_gameover);
        EzDrawStringB(320-16/2*8, 320, buf);
        */
    }

    draw();
}


void draw(void){
    int i, ix;
    char buf[32];

    draw_ground();
    draw_sensor();
    draw_info();
    draw_ship_shadows();
    if (myship.state != ss_dead){
        draw_myship();
    }
    for (i = 0 ; i < MYBULLET_MAX ; i++){
        if (mybullet[i].state != bs_dead){
            draw_mybullet(&mybullet[i]);
        }
    }
    for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
        if (enemyship[i].state != ss_dead){
            draw_enemyship(&enemyship[i]);
        }
    }
    for (i = 0 ; i < ENEMYBULLET_MAX ; i++){
        if (enemybullet[i].state != bs_dead){
            draw_enemybullet(&enemybullet[i]);
        }
    }
    if (current_group.type == et_dragon){
        for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
            if (enemyship[i].state != ss_dead && enemyship[i].type == et_dragon_head){
                draw_neck(&enemyship[i]);
                if (enemyship[i].action == ea_dragon_head_breath){
                    draw_breath(&enemyship[i]);
                    check_hit_breath(&enemyship[i]);
                }
            }
        }
    }
    else if (current_group.type == et_core){
        draw_lasers();
        check_hit_lasers();
    }
    if (is_pause){
        //        SetColor(7);
        set_color(ct_info);
        EzDrawStringB(320-9/2*8, 320, "P A U S E");
    }
    if (myship.state == ss_dead){
        if (gamestate_counter < 17){
            ix = gamestate_counter;
        }
        else {
            ix = 17;
        }
        strncpy(buf, "G A M E  O V E R", ix);
        buf[ix] = '\0';
        //        SetColor(2);
        set_color(ct_gameover);
        EzDrawStringB(320-16/2*8, 320, buf);
    }    
    EzShowBuffer();
}

void draw_info(void){
  char buf[256];

  //  SetColor(7);
  set_color(ct_info);
  sprintf(buf, "HI-SCORE %08d", 
          achievements.top_score[config.options[ci_stage]]);
  EzDrawStringB(info_lt_x, info_lt_y, buf);
  sprintf(buf, "SCORE    %08d", score);
  EzDrawStringB(info_lt_x, info_lt_y + INFO_LINESKIP, buf);
  sprintf(buf, "STAGE %03d LEFT %02d", stage, myship.left);
  EzDrawStringB(info_lt_x, info_lt_y + INFO_LINESKIP*2, buf);
}

void set_view(view_mode_t new_view_mode){
    set_view_mode(new_view_mode);
    switch (new_view_mode){
    case vm_bird:
        info_lt_x = INFO_LT_X;
        info_lt_y = INFO_LT_Y;
        sensor_lt_x = SENSOR_LT_X;
        sensor_lt_y = SENSOR_LT_Y;
        break;
    case vm_top:
        info_lt_x = INFO_LT_X_TOP;
        info_lt_y = INFO_LT_Y_TOP;
        sensor_lt_x = SENSOR_LT_X_TOP;
        sensor_lt_y = SENSOR_LT_Y_TOP;
        break;
    case vm_isom:
        info_lt_x = INFO_LT_X_ISO;
        info_lt_y = INFO_LT_Y_ISO;
        sensor_lt_x = SENSOR_LT_X_ISO;
        sensor_lt_y = SENSOR_LT_Y_ISO;
        break;
    default:
        assert(0);
    }

    /*
    sensor_lt_y = SENSOR_U_LT_Y;
        sensor_rb_y = SENSOR_U_RB_Y;
    }
    else {
        sensor_lt_y = SENSOR_LT_Y;
        sensor_rb_y = SENSOR_RB_Y;
    }
    */
}

void toggle_view_mode(void){
    view_mode_t view_mode;
    //    view_mode = get_view_mode()+1;
    view_mode = config.options[ci_view] + 1;
    if (view_mode == vm_end) view_mode = vm_begin;
    set_view(view_mode);
    config.options[ci_view] = view_mode;
}

void draw_ground(void){
    int x, z;
    point3d_t st3d, en3d;

    set_color(ct_background);
    EzFillBoxB(0, 0, WIN_WIDTH, WIN_HEIGHT);
    // 一時停止中はアニメーションを止める
    if (!is_pause){
        grid_counter --;
        if (grid_counter == -1){
            grid_counter = 3;
        }
    }

    if (set_color(ct_ground)) return;
    st3d.x = XMIN;
    st3d.y = GROUND_HEIGHT;
    en3d.x = XMAX;
    en3d.y = GROUND_HEIGHT;
    //    SetGroundColor();

    for (z = ZMIN ; z <= ZMAX ; z += GRIDSIZE){
        st3d.z = z;
        en3d.z = z;
        if (z/GRIDSIZE % 4 == grid_counter){
            draw_3dline(st3d, en3d);
        }
        else {
        }
    } 
    //    SetGroundColor();
    st3d.z = ZMIN;
    st3d.y = GROUND_HEIGHT;
    en3d.z = ZMAX;
    en3d.y = GROUND_HEIGHT;
    for (x = XMIN ; x <=XMAX ; x += GRIDSIZE*4){ 
        st3d.x = x;
        en3d.x = x;
        draw_3dline(st3d, en3d);
    }
}


#if 0
void draw_ground_topview(void){
    int x, z;
    point3d_t st3d, en3d;

    set_color(ct_background);
    EzFillBoxB(0, 0, WIN_WIDTH, WIN_HEIGHT);
    // 一時停止中はアニメーションを止める
    if (!is_pause){
        grid_counter --;
        if (grid_counter == -1){
            grid_counter = 3;
        }
    }

    if (set_color(ct_ground)) return;
    st3d.x = XMIN;
    st3d.y = GROUND_HEIGHT;
    en3d.x = XMAX;
    en3d.y = GROUND_HEIGHT;
    //    SetGroundColor();

    for (z = ZMIN ; z <= ZMAX ; z += GRIDSIZE){
        st3d.z = z;
        en3d.z = z;
        if (z/GRIDSIZE % 4 == grid_counter){
            draw_3dline(st3d, en3d);
        }
        else {
        }
    } 
    //    SetGroundColor();
    st3d.z = ZMIN;
    st3d.y = GROUND_HEIGHT;
    en3d.z = ZMAX;
    en3d.y = GROUND_HEIGHT;
    for (x = XMIN ; x <=XMAX ; x += GRIDSIZE*4){ 
        st3d.x = x;
        en3d.x = x;
        draw_3dline(st3d, en3d);
    }
}
#endif

void animate_sensor(void){
    if (!IsEnemyTypeBoss(current_group.type)) {
        animate_sensor_model(&enemyship_model[current_group.type]);
    }
}

void draw_sensor(void){
    char buf[256];
    static int blink = 0;

    if (grid_counter == 0){
        if (blink){
            blink = 0;
        }
        else {
            blink = 1;
        }
    }

    EzDrawBoxB(sensor_lt_x, sensor_lt_y,
               sensor_lt_x + SENSOR_WIDTH, sensor_lt_y + SENSOR_HEIGHT);
    sprintf(buf, "<%s> x %02d",
            enemydata[current_group.type].name,
            current_group.enemy_num + current_group_alive_enemy_num);
    //    SetColor(7);
    set_color(ct_info);
    EzDrawStringB(sensor_lt_x + 4, sensor_lt_y + 16, "ENEMY DETECTED");
    EzDrawStringB(sensor_lt_x + 4, sensor_lt_y + 32, buf);

    if (IsEnemyTypeBoss(current_group.type)) {
        if (blink){
            //            SetColor(2);
            set_color(ct_warn);
            EzDrawStringB(sensor_lt_x + 24, sensor_lt_y + 80,
                              "* UNKNOWN *");	
        }
    }
    else {
        draw_sensor_model(&enemyship_model[current_group.type]);
    }
}

void draw_ship_shadows(void){
    int i;
    enemyship_t *ship;

    if (myship.state == ss_alive){
        draw_shadow(&myship_model, myship.xa+1, myship.pos);
    }
    for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
        ship = &enemyship[i];
        if (ship->state == ss_alive){
            draw_shadow(&enemyship_model[ship->type], ship->motion, ship->pos);
        }
    }
}

void draw_myship(void){
    if (myship.state == ss_alive){
        draw_model(&myship_model, myship.xa+1, myship.pos);
        if (ShowHitBox()) {
            draw_hitbox(&myship_model, myship.pos);
        }
    }
    else {
        //    printf("explode %d (%d, %d)\n", myship.explode_motion,
        //           myship.pos.x, myship.pos.z);
        // 爆発はエリア外に出る可能性があるので、
        // その場合は描画をスキップする
        if (myship.pos.x > XMIN + 0 && myship.pos.x < XMAX - 0 &&
            myship.pos.z > ZMIN + 0 && myship.pos.z < ZMAX - 0)
        {
            draw_explosion(&myship_model, myship.xa+1, myship.explode_motion,
                           myship.pos);
        }
    }
}

void draw_mybullet(bullet_t *bullet){
    point3d_t h3d, l3d, r3d;

    h3d.x = bullet->pos.x;
    h3d.y = FLIGHT_HEIGHT;
    h3d.z = bullet->pos.z + 8;
    l3d.x = bullet->pos.x - 4;
    l3d.y = FLIGHT_HEIGHT;
    l3d.z = bullet->pos.z - 8;
    r3d.x = bullet->pos.x + 4;
    r3d.y = FLIGHT_HEIGHT;
    r3d.z = bullet->pos.z - 8;

    //    SetColor(6);
    set_color(ct_mybullet);
    draw_3dline(h3d, l3d);
    draw_3dline(h3d, r3d);
}

void draw_enemyship(enemyship_t *ship){
    if (ship->state == ss_alive){
        //	printf("draw motion:%d\n", ship->motion);
        draw_model(&enemyship_model[ship->type], ship->motion, ship->pos);
        if (ShowHitBox() && ship->can_hit){
            draw_hitbox(&enemyship_model[ship->type], ship->pos);
        }
    }
    else { // ss_explodingの処理
        //        printf("draw explosion %d %d %d\n", ship->type, ship->motion, ship->explode_motion);
        draw_explosion(&enemyship_model[ship->type],
                       ship->motion, ship->explode_motion, 
                       ship->pos);	
    }
}


void draw_enemybullet(bullet_t *bullet){
    point3d_t h3d, l3d, r3d;

    h3d.x = bullet->pos.x;
    h3d.y = FLIGHT_HEIGHT;
    h3d.z = bullet->pos.z - 8;
    l3d.x = bullet->pos.x - 4;
    l3d.y = FLIGHT_HEIGHT;
    l3d.z = bullet->pos.z + 8;
    r3d.x = bullet->pos.x + 4;
    r3d.y = FLIGHT_HEIGHT;
    r3d.z = bullet->pos.z + 8;

    //    SetColor(2);
    set_color(ct_enemybullet);
    draw_3dline(h3d, l3d);
    draw_3dline(h3d, r3d);
}

void draw_neck(enemyship_t *head){
    point3d_t b3d, t3d;

    b3d.x = dragon->pos.x;
    b3d.y = dragon->pos.y - 90;
    b3d.z = dragon->pos.z;

    t3d.x = head->pos.x;
    t3d.y = FLIGHT_HEIGHT;
    t3d.z = head->pos.z;

    //    SetColor(4);
    set_color(ct_enemyship);
    draw_3dline(b3d, t3d);
}

void draw_breath(enemyship_t *head){
    int i, dir;
    point3d_t top;

    dir = head->motion / 6;
    //    SetColor(2);
    set_color(ct_enemybullet);
    for (i = 0 ; i < head->fire_counter ; i++){
        top = add_vector3d(breath[dir][i].t, head->pos);
        if (top.x < XMIN || top.x > XMAX || top.z < ZMIN - 96){
            break;
        }
        draw_3dline(top,
                    add_vector3d(breath[dir][i].r, head->pos));
        draw_3dline(add_vector3d(breath[dir][i].r, head->pos),
                    add_vector3d(breath[dir][i].b, head->pos));
        draw_3dline(add_vector3d(breath[dir][i].b, head->pos),
                    add_vector3d(breath[dir][i].l, head->pos));
        draw_3dline(add_vector3d(breath[dir][i].l, head->pos),
                    top);
    }
}

void check_hit_breath(enemyship_t *head){
    int i, dir;
    point3d_t top;

    dir = head->motion / 6;
    for (i = 0 ; i < head->fire_counter ; i++){
        top = add_vector3d(breath[dir][i].t, head->pos);
        if (top.x < XMIN || top.x > XMAX || top.z < ZMIN - 96){
            break;
        }
        if (is_hit(top, myship.pos, myship_model.xsize, myship_model.zsize) &&
            myship.state == ss_alive &&
            immortal_counter == 0 && 
            is_mortal == 1)
        {
            myship.state = ss_exploding;
            myship.explode_motion = 0;
            if (!is_cheat){
                IncUInt(achievements.total_shotdown_type[et_dragon]);
            }
        }
    }
}

void draw_lasers(void){
    int i;
    //    SetColor(6);
    set_color(ct_laser);
    for (i = 0 ; i < LASER_MAX_NUM ; i++){
        if (lasers[i].is_valid){
            draw_laser(&lasers[i]);
        }
    }
}

void draw_laser(laser_t *laser){
    point3d_t st, en;
    int i;

    //    i = (laser->head - 1 + LAZER_MAX_LEN) % LASER_MAX_LEN;
    i = laser->head;
    st = laser->nodes[i].t;
    IncLaserIx(i);
    do {
        en = laser->nodes[i].t;
        draw_3dline(st, en);
        /*
        printf("draw laser (%d, %d, %d)-(%d, %d, %d)\n", 
               st.x, st.y, st.z,
               en.x, en.y, en.z
               );
        */
        st = en;
        IncLaserIx(i);
    } while (i != SuccLaserIx(laser->tail));
}


void move_lasers(void){
    int i;
    for (i = 0 ; i < LASER_MAX_NUM ; i++){
        if (lasers[i].is_valid){
            move_laser(&lasers[i]);
        }
    }
}

void check_hit_lasers(void){
    int i;
    for (i = 0 ; i < LASER_MAX_NUM ; i++){
        if (lasers[i].is_valid){
            check_hit_laser(&lasers[i]);
        }
    }
}

void check_hit_laser(laser_t *laser){
    point3d_t st, en, pos;
    int i, xl, zl, dist, j, n;

    i = laser->head;
    st = laser->nodes[i].t;
    IncLaserIx(i);
    do {
        en = laser->nodes[i].t;
        xl = abs(st.x-en.x);
        zl = abs(st.z-en.z);
        dist = xl > zl ? xl : zl;
        n = dist / myship_model.zsize+1;
        pos.y = st.y;
        for (j = 0 ; j <= n ; j++){
            pos.x = st.x + (en.x-st.x)*j/n;
            pos.z = st.z + (en.z-st.z)*j/n;
            if (is_hit(pos, myship.pos,
                       myship_model.xsize, myship_model.zsize) &&
                myship.state == ss_alive &&
                immortal_counter == 0 && 
                is_mortal == 1)
            {
                myship.state = ss_exploding;
                myship.explode_motion = 0;
                if (!is_cheat){
                    IncUInt(achievements.total_shotdown_type[et_core]);
                }
            }
        }
        st = en;
        IncLaserIx(i);
    } while (i != SuccLaserIx(laser->tail));
}

void move_laser(laser_t *laser){
    point3d_t new_pos; //, head_pos;
    int new_tail = SuccLaserIx(laser->tail);
    int rx, rz, xa, za;

    new_pos = laser->nodes[laser->tail].t;
    new_pos.x += laser->vx;
    new_pos.z += laser->vz;
    // 画面外に出たらビームはそれ以上伸びない
    // 画面外なのでMIN/MAXよりさらに少し外側
    if (new_pos.x < XMIN-128 || new_pos.x > XMAX+128 ||
        new_pos.z < ZMIN-128 || new_pos.z > ZMAX+128)
    {
        IncLaserIx(laser->head);
//        head_pos = laser->nodes[laser->head].t;
        if (laser->head == laser->tail){
            laser->is_valid = 0;
        }
    //        if (head_pos.x < XMIN-128 || head_pos.x > XMAX+128 ||
    //            head_pos.z < ZMIN-128 || head_pos.z > ZMAX+128)
    //        {    
    //            laser->is_valid = 0;
    //        }
        return;
    }
    laser->nodes[new_tail].t = new_pos;
    laser->tail = new_tail;
    if (laser->tail == laser->head){
        IncLaserIx(laser->head);
    }

    // 自機が破壊されたら、レーザのホーミングを行わない
    // さもないと復活中の自機の上をレーザが周回することがある
    if (myship.state != ss_alive) return;
    rx = laser->nodes[laser->tail].t.x - myship.pos.x;
    rz = laser->nodes[laser->tail].t.z - myship.pos.z;
    if (abs(rx) > abs(rz)){
        xa = LASER_ACCERATION;
        za = LASER_ACCERATION/2;
    }
    else {
        xa = LASER_ACCERATION/2;
        za = LASER_ACCERATION;
    }

    if (rx < 0) {
        laser->vx += xa;
    }
    else if (rx > 0){
        laser->vx -= xa;
    }
    if (rz < 0) {
        laser->vz += za;
    }
    else if (rz > 0){
        laser->vz -= za;
    }
}

void move_myship(void){
  if (immortal_counter){
	immortal_counter --;
    //	printf("IC:%d\n", immortal_counter);
	if (immortal_counter % 4){
	  myship_model.color_type = ct_myship_blink;
	}
	else {
	  myship_model.color_type = ct_myship;
	}
	if (immortal_counter == 0){
        myship_model.color_type = ct_myship;
	}
  }  
  if (myship.state == ss_alive){
      // 現在の方向キー押下状況から移動方向を決定
      dir = dir_table[dir_keymap];
      // ただし、前回のタイマ割り込みから今回までにキーを一瞬押して離した場合、
      // 押したことが認識されない
      // 直前のキー押下時点での方向はdir_prevに入っているので、
      // 現在方向無入力状態でもdir_prevがいずれかの方向を指していればそちらを採用する
      if (dir == 0 && dir_prev != 0){
          dir = dir_prev;
      }
      // 採用の有無に関わりなく、dir_prevはタイマ割り込み毎に無効化する
      dir_prev = 0;
      myship.xa = move_vec[dir].x;
      myship.za = move_vec[dir].y;
  }

  myship.pos.x += myship.xa * MYSHIP_SPEED;
  myship.pos.z += myship.za * MYSHIP_SPEED;
  if (myship.state == ss_alive){
	if (myship.pos.x < XMIN + 16 || myship.pos.x > XMAX - 16){
	  myship.pos.x -= myship.xa * MYSHIP_SPEED;
      // 端を超えて進もうとしていると爆発エフェクトがすぐに
      // エリア外に出て終わってしまうので、
      // こっそり速度を0にしておきたいが、
      // x方向移動はモーションが変わるのでやめておく。
      // クリッピングを使ってエリア外でも爆発は描画すべきか。
      // myship.xa = 0;
	}
	if (myship.pos.z < ZMIN + 32 || myship.pos.z > ZMAX - 256){
	  myship.pos.z -= myship.za * MYSHIP_SPEED;
      // 端を超えて進もうとしていると爆発エフェクトがすぐに
      // エリア外に出て終わってしまうので、
      // こっそり速度を0にしておく。
      // z方向移動はモーションが変わらないから気づかれない
      myship.za = 0;
	}
	if (is_autofire || is_fire){
	  fire();
	}
  }
  else if (myship.state == ss_exploding){
	myship.explode_motion ++;
    if (myship.left == 0){
        game_speed += GAME_SPEED_SLOWDOWN;
    }
	if (myship.explode_motion == MYSHIP_EXPLODE_MAX){
        //		myship.pos.x < XMIN + 0 || myship.pos.x > XMAX - 0 ||
        //		myship.pos.z < ZMIN + 0 || myship.pos.z > ZMAX - 0){
	  if (myship.left == 0){
		myship.state = ss_dead;
	  }
	  else {
		myship.left --;
		immortal_counter = IMMORTAL_TIME;
		init_myship();
	  }
	}
  }
}

void move_mybullet(bullet_t *bullet){
  bullet->pos.z += MYBULLET_SPEED;
  if (bullet->pos.z > ZMAX){
	bullet->state = bs_dead;
  }
}

void check_hit_enemyship(enemyship_t *ship){
    int i, no;
    enemytype_t killed_type = et_invalid;
    //    printf("%d(%d, %d) (%dx%d)\n", ship->can_hit,
    //           ship->pos.x, ship->pos.z,
    //enemyship_model[ship->type].xsize,
    //enemyship_model[ship->type].zsize);

    // BOSSのデフォルト色を4にする
    // 他の敵でも毎回やるのは無駄だが・・・
    enemyship_model[ship->type].color_type = ct_enemyship;
    if (is_hit(ship->pos, myship.pos, 
               enemyship_model[ship->type].xsize + myship_model.xsize,
               enemyship_model[ship->type].zsize + myship_model.zsize) &&
        myship.state == ss_alive &&
        immortal_counter == 0 && is_mortal == 1 &&
        ship->can_hit != 0)
    {
        myship.state = ss_exploding;
        myship.explode_motion = 0;
        if (!is_cheat){
            if (ship->type == et_dragon_head){
                IncUInt(achievements.total_crush_type[et_dragon]);
            }
            else {
                IncUInt(achievements.total_crush_type[ship->type]);
            }
        }
    }

    for (i = 0 ; i < MYBULLET_MAX ; i++){
        if (mybullet[i].state == bs_alive){
            if (ship->can_hit &&
                is_hit(ship->pos, mybullet[i].pos, 
                       enemyship_model[ship->type].xsize + MYBULLET_SIZE,
                       enemyship_model[ship->type].zsize + //MYBULLET_SIZE +
                       MYBULLET_SPEED))
            {
                mybullet[i].state = bs_dead;
                if (IsEnemyBlockingBullet(ship)){
                    // WALLの場合は命中している弾すべてを無効化するため、
                    // ループ最後まで実行する。
                    printf("block!\n");
                }
                else if (IsEnemyTypeBoss(ship->type)) {
                    enemyship_model[ship->type].color_type = ct_enemyship_hit;
                    boss_life --;
                    printf("boss life:%d/%d\n", boss_life, boss_life_max);

                    if (boss_life <= 0){
                        // DRAGON戦の場合、本体に当たり判定はないので
                        // dragon_headに対する処理
                        if (ship->type == et_dragon_head){
                            dragon_head_num --;
                            printf("head destroy left:%d\n",
                                   dragon_head_num);
                            // 最後の頭を撃破したら本体も撃破->CORE登場
                            if (dragon_head_num == 0){
                                printf("create core\n");
                                dragon->state = ss_exploding;
                                no = allocate_enemyship();
                                enemyship[no].type = et_core;
                                create_core(&enemyship[no]);
                                dragon = NULL;
                                // レーダに名前を表示するため
                                current_group.type = et_core;
                                killed_type = et_dragon;
                                /*
                                // 通常のプレイなら撃破数加算
                                if (!is_cheat){
                                    IncUInt(achievements.total_kill_type[et_dragon]);
                                }
                                add_score(enemydata[et_dragon].score);
                                */
                            }
                            // 頭が残っていたらライフを元に戻す
                            else {
                                init_boss_life(DRAGON_BASELIFE, DRAGON_ADDLIFE);
                                killed_type = et_invalid;
                            }
                        }
                        else {
                            killed_type = ship->type;
                        }
                        // 実際に破壊した(ことにする)敵について撃破数とスコア加算
                        // 同一ターンである弾に当たって敵HP==0になったのに
                        // 別の弾にも当たって二重加算されるのを防ぐため、
                        // ss_aliveであるかチェックする
                        if (killed_type != et_invalid &&
                            ship->state == ss_alive)
                        {
                            // 通常のプレイなら撃破数加算
                            if (!is_cheat){
                                IncUInt(achievements.total_kill_type[killed_type]);
                            }
                            add_score(enemydata[killed_type].score);
                        }
                        ship->state = ss_exploding;
                        /*
                        else if (ship->type == et_core){
                            // 通常のプレイなら撃破数加算
                            if (!is_cheat){
                                IncUInt(achievements.total_kill_type[et_core]);
                            }
                            add_score(enemydata[et_core].score);
                        }
                        else {  // DRAGON以外のボス
                            // 通常のプレイなら撃破数加算
                            if (!is_cheat){
                                printf("kill boss %d %d\n", ship->type, achievements.total_kill_type[ship->type]);
                                IncUInt(achievements.total_kill_type[ship->type]);
                            }
                            add_score(enemydata[ship->type].score);
                        }
                        */
                    }
                }
                else {
                    ship->state = ss_exploding;
                    // 通常のプレイなら撃破数加算
                    if (!is_cheat){
                        IncUInt(achievements.total_kill_type[ship->type]);
                    }
                    add_score(enemydata[ship->type].score);
                    mybullet[i].state = bs_dead;
                    return;
                }
            }
        }
    }
    // WALLの場合は敵の弾も無効化
    //    if (ship->type == et_wall){
    if (IsEnemyBlockingBullet(ship)){
        for (i = 0 ; i < ENEMYBULLET_MAX ; i++){
            if (enemybullet[i].state == bs_alive){
                if (is_hit(ship->pos, enemybullet[i].pos, 
                           enemyship_model[ship->type].xsize +
                           ENEMYBULLET_SIZE,
                           enemyship_model[ship->type].zsize +//MYBULLET_SIZE +
                           abs(enemybullet[i].za)))
                {
                    enemybullet[i].state = bs_dead;
                }
            }
        }
    }
}

void check_hit_enemybullet(bullet_t *bullet){
    if (is_hit(bullet->pos, myship.pos, 
               myship_model.xsize, myship_model.zsize) &&
        myship.state == ss_alive && 
        immortal_counter == 0 &&
        is_mortal == 1)
    {
        myship.state = ss_exploding;
        myship.explode_motion = 0;
        if (!is_cheat){
            IncUInt(achievements.total_shotdown_type[bullet->owner_type]);
        }
	bullet->state = bs_dead;
  }
}

void create_enemyship(void){
    int no;
    enemyship_t *ship;
    // 前回の敵機生成で現ステージの機数を使い切っていたら、
    // ステージを進めてから今回の敵機を生成

    //    printf("group_num %d %d current_group num:%d type:%s\n", group_num, group_no, current_group.enemy_num, enemydata[current_group.type].name);

    if (current_group.enemy_num == 0){
        group_no ++;
        if (group_no >= group_num){
            group_no = 0;
            set_stage(stage+1);
        }
        //        printf("group %d\n", group_no);
        current_group = group[group_no];
        current_group_alive_enemy_num = 0;
    }
//    printf("create enemy type:%d\n", current_group.type);
    no = allocate_enemyship();
    if (no != -1){
        ship =  &enemyship[no];
        switch (ship->type) {
        case et_scout:
            create_scout(ship);
            break;
        case et_wall:
            create_wall(ship);
            break;
        case et_drone:
            create_drone(ship);
            break;
        case et_fighter:
            create_fighter(ship);
            break;
        case et_copter:
            create_copter(ship);
            break;
        case et_lancer:
            create_lancer(ship);
            break;
        case et_walker:
            create_walker(ship);
            break;
        case et_turret:
            create_turret(ship);
            break;
        case et_fortress:
            create_fortress(ship);
            break;
        case et_bomber:
            create_bomber(ship);
            break;
        case et_wyvern:
            create_wyvern(ship);
            break;
        case et_byte:
            create_byte(ship);
            break;
        case et_gunroid:
            create_gunroid(ship);
            break;
        case et_dragon_head:
            create_dragon_head(ship);
            break;
        case et_dragon:
            create_dragon(ship);
            break;
        case et_core:
            create_core(ship);
            break;
        default:
            perror("undefined ship type in create_enemyship");
        }
        ship->group_no = group_no;
        current_group.enemy_num --;
        current_group_alive_enemy_num ++;
        printf("%s created (%d left).\n", 
               enemydata[ship->type].name, current_group.enemy_num);
        return;
    }
}

// enemyship[]から空きを見つけて初期化し、インデックスを返す
// 空きがない場合は-1を返す
int allocate_enemyship(void){
    int i;
    enemyship_t *ship;

    for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
        if (enemyship[i].state == ss_dead){
            enemyship_num ++;
            //            printf("create enemyship no %d... now %d enemies\n", i, enemyship_num);
            ship =  &enemyship[i];
            ship->state = ss_alive;
            ship->type = current_group.type;
            ship->motion = 0;
            ship->explode_motion = 0;
            ship->fire_counter = 0;
            ship->counter = 0;
            ship->action = ea_init;
            ship->can_hit = -1;
            return i;
        }
    }
    return -1;
}


int find_alive_enemyship(enemytype_t type){
    int i, no;
    // とりあえずランダムに選択してみて生きていればそれを返す
    for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
        no = RRnd(0, ENEMYSHIP_MAX-1);
        if (enemyship[no].state == ss_alive && enemyship[no].type == type){
            return no;
        }
    }
    // 何回かやって見つからなければ最初から順に探す
    for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
        if (enemyship[no].state == ss_alive && enemyship[no].type == type){
            return i;
        }
    }
    // 見つからなければ-1を返す
    return -1;
}

void delete_enemyship(enemyship_t *ship){
  ship->state = ss_dead;
  enemyship_num --;
  if (ship->group_no == group_no){
      current_group_alive_enemy_num --;
  }
  printf("delete enemyship... now %d enemies\n", enemyship_num);
}


void create_scout(enemyship_t *ship){
  ship->pos.x = RRnd(XMIN+100,XMAX-100);
  ship->pos.y = FLIGHT_HEIGHT;
  ship->pos.z = ZMAX - 64;
  //  printf("%d %d %d\n", ship->pos.x, ship->pos.y, ship->pos.z);
  if (RRnd(0,1) == 0){
	ship->xspeed = -SCOUT_XSPEED;
  }
  else {
	ship->xspeed = SCOUT_XSPEED;
  }
  ship->zspeed = -SCOUT_ZSPEED;
}


void create_wall(enemyship_t *ship){
  ship->pos.x = RRnd(XMIN+32,XMAX-32);
  ship->pos.y = FLIGHT_HEIGHT;
  ship->pos.z = ZMAX - 64;
  ship->xspeed = 0;
  ship->zspeed = -WALL_ZSPEED;
}

void create_drone(enemyship_t *ship){
  if (RRnd(0, 1)==0){
	ship->pos.x = RRnd(XMIN+32,XMIN/2);
  }
  else {
	ship->pos.x = RRnd(XMAX/2,XMAX-32);
  }
  ship->pos.y = FLIGHT_HEIGHT;
  ship->pos.z = ZMAX - 64;
  ship->xspeed = 0;
  ship->zspeed = -DRONE_ZSPEED;
  ship->fire_rate = DRONE_FIRE_RATE;
}

void create_fighter(enemyship_t *ship){
  ship->pos.x = RRnd(XMIN+64,XMAX-64);
  ship->pos.y = FLIGHT_HEIGHT;
  ship->pos.z = ZMAX - 64;
  ship->xspeed = 0;
  ship->zspeed = -FIGHTER_ZSPEED;
}

void create_copter(enemyship_t *ship){
  if (RRnd(0, 1)==0){
	ship->pos.x = RRnd(XMIN+32,XMIN/2);
	ship->xspeed = 1;
  }
  else {
	ship->pos.x = RRnd(XMAX/2,XMAX-32);
	ship->xspeed = -1;
  }
  // 発射頻度が低いので、一発目は早めに撃つ
  ship->fire_rate = COPTER_FIRE_RATE / 2 + RRnd(-6, 6);
  ship->pos.y = FLIGHT_HEIGHT;
  ship->pos.z = ZMAX - 64;
  ship->zspeed = -COPTER_ZSPEED+1;
}

void create_lancer(enemyship_t *ship){
  ship->pos.x = RRnd(XMIN+64,XMAX-64);
  ship->pos.y = FLIGHT_HEIGHT;
  ship->pos.z = ZMAX - 64;
  ship->xspeed = 0;
  ship->zspeed = -LANCER_ZSPEED;
}

void create_walker(enemyship_t *ship){
  if (RRnd(0, 1)==0){
	ship->pos.x = XMAX-32;
	ship->xspeed = -WALKER_XSPEED;
	ship->motion = 0;
  }
  else {
	ship->pos.x = XMIN+32;
	ship->xspeed = WALKER_XSPEED;
	ship->motion = 16;
  }
  ship->pos.z = ZMAX - RRnd(64, 192);
  ship->zspeed = -WALKER_ZSPEED;
  ship->fire_rate = WALKER_FIRE_RATE + RRnd(-6, 6);
  ship->pos.y = GROUND_HEIGHT;
}


void create_turret(enemyship_t *ship){
  ship->pos.x = RRnd(XMIN+32,XMAX-32);
  ship->pos.y = GROUND_HEIGHT;
  ship->pos.z = ZMAX - 64;
  ship->xspeed = 0;
  ship->zspeed = -GRIDSIZE;
  ship->fire_rate = TURRET_FIRE_RATE;
  ship->fire_counter = TURRET_FIRE_RATE-1;
}

void create_fortress(enemyship_t *ship){
  ship->pos.x = 0;
  ship->pos.y = (GROUND_HEIGHT + FLIGHT_HEIGHT) / 2;
  ship->pos.z = ZMAX - 128;
  ship->xspeed = FORTRESS_SPEED;
  ship->zspeed = -FORTRESS_SPEED;
  ship->fire_rate = FORTRESS_FIRE_RATE + RRnd(-6, 6);
  init_boss_life(FORTRESS_BASELIFE, FORTRESS_ADDLIFE);
}


void create_bomber(enemyship_t *ship){
    ship->pos.x = RRnd(-200, 200);
    ship->pos.y = GROUND_HEIGHT;
    ship->pos.z = ZMAX - 128;
    ship->xspeed = 0;
    ship->zspeed = -GRIDSIZE;
    ship->fire_rate = BOMBER_FIRE_RATE + RRnd(-6, 6);
    init_boss_life(BOMBER_BASELIFE, BOMBER_ADDLIFE);
    ship->can_hit = 0; // 離陸時は当たり判定なし
}

void create_wyvern(enemyship_t *ship){
    ship->pos.x = myship.pos.x;
    ship->pos.y = FLIGHT_HEIGHT - 256;
    ship->pos.z = ZMIN-64;
    ship->xspeed = -8;
    ship->zspeed = GRIDSIZE; // 登場時は低速で移動
    ship->fire_rate = WYVERN_FIRE_RATE + RRnd(-6, 6);
    init_boss_life(WYVERN_BASELIFE, WYVERN_ADDLIFE);
    ship->can_hit = 0; // 登場時は当たり判定なし
}

void create_byte(enemyship_t *ship){
    int i, no;

    for (i = 0 ; i < BYTE_NUM ; i++){
        // 最初のshipは呼び出し前に確保済み。あとは毎回ここで確保する
        if (i > 0){
            printf("create byte\n");
            no = allocate_enemyship();
            if (no != -1){
                ship =  &enemyship[no];
                ship->type = et_byte;
                current_group_alive_enemy_num ++;
            }
            else {
                return;
            }
        }
        printf("create byte no%d %d\n", i, enemyship_num);
        if (current_group.type == et_byte){
            ship->pos.x = (i-BYTE_NUM/2)*40;
            ship->pos.y = GROUND_HEIGHT;
            ship->pos.z = ZMAX;
            ship->xspeed = 0;
            ship->zspeed = -GRIDSIZE;
            ship->motion = 8;
            ship->can_hit = 0; // 登場時は当たり判定なし
        }
        else {
            ship->pos.x = 0;
            ship->pos.y = FLIGHT_HEIGHT;
            ship->pos.z = BYTE_SHIELD_ZPOS + 80 + i*10;
            ship->xspeed = 0;
            ship->zspeed = 0;
        }
    }
}

void create_gunroid(enemyship_t *ship){
    ship->pos.x = 0;
    ship->pos.y = (FLIGHT_HEIGHT + GROUND_HEIGHT)/2;
    ship->pos.z = ZMAX;
    //    ship->pos.z = ZMAX-700; // !!
    //    ship->pos.x = -400; // !!
    ship->xspeed = -8;
    ship->zspeed = 16;
    //    ship->fire_rate = WYVERN_FIRE_RATE + RRnd(-6, 6);
    init_boss_life(GUNROID_BASELIFE, GUNROID_ADDLIFE);
    ship->can_hit = 0; // 登場時は当たり判定なし
}

void create_dragon_head(enemyship_t *ship){
    ship->pos.x = RRnd(-DRAGON_HEAD_MOVE_RANGE, DRAGON_HEAD_MOVE_RANGE);
    ship->pos.y = FLIGHT_HEIGHT + 20;
    ship->pos.z = ZMAX;
    ship->xspeed = 0;
    ship->zspeed = 0;
    ship->motion = 12;
    ship->can_hit = 0; // 登場時は当たり判定なし
}

void create_dragon(enemyship_t *ship){
    int i, no;
    enemyship_t *head;

    dragon = ship;
    ship->pos.x = 0;
    ship->pos.y = GROUND_HEIGHT - 16;
    //    ship->pos.z = DRAGON_ZPOS;
    ship->pos.z = ZMAX;
    ship->zspeed = -GRIDSIZE;
    dragon_head_num = DRAGON_HEAD_BASE_NUM + (stage-1) / 3;
    if (dragon_head_num > DRAGON_HEAD_MAX_NUM){
        dragon_head_num = DRAGON_HEAD_MAX_NUM;
    }
    init_boss_life(DRAGON_BASELIFE, DRAGON_ADDLIFE);
    ship->can_hit = 0; // 本体には当たり判定なし
    for (i = 0 ; i < dragon_head_num ; i++){
        no = allocate_enemyship();
        if (no != -1){
            head =  &enemyship[no];
            head->type = et_dragon_head;
            create_dragon_head(head);
        }
    }
}

void create_core(enemyship_t *ship){
    int i;

    for (i = 0 ; i < LASER_MAX_NUM ; i++){
        lasers[i].is_valid = 0;
    }
    //    ship->pos.x = RRnd(-300, 300);
    //    ship->pos.x = 00;
    //    ship->pos.y = FLIGHT_HEIGHT + 20;
    //    ship->pos.y = FLIGHT_HEIGHT + -320;
    //    ship->pos.z = DRAGON_ZPOS-80;
    ship->pos = dragon->pos;
    ship->pos.y -= 40;  // 上下位置の調整
    ship->xspeed = -8;
    //    ship->zspeed = -8;
    ship->motion = 0;
    ship->can_hit = 0; // 登場時は当たり判定なし
    init_boss_life(CORE_BASELIFE, CORE_ADDLIFE);
}


void move_enemyship(enemyship_t *ship, int no){
    int i;
    if (ship->state == ss_alive){
        switch (ship->type){
        case et_scout:
            move_scout(ship);
            break;
        case et_wall:
            move_wall(ship);
            break;
        case et_drone:
            move_drone(ship);
            break;
        case et_fighter:
            move_fighter(ship);
            break;
        case et_copter:
            move_copter(ship);
            break;
        case et_lancer:
            move_lancer(ship);
            break;
        case et_walker:
            move_walker(ship);
            break;
        case et_turret:
            move_turret(ship);
            break;
        case et_fortress:
            move_fortress(ship);
            break;
        case et_bomber:
            move_bomber(ship);
            break;
        case et_wyvern:
            move_wyvern(ship);
            break;
        case et_byte:
            move_byte(ship, no);
            break;
        case et_gunroid:
            move_gunroid(ship);
            break;
        case et_dragon_head:
            move_dragon_head(ship);
            break;
        case et_dragon:
            move_dragon(ship);
            break;
        case et_core:
            move_core(ship);
            break;
        default:
            perror("unknown ship type in move_enemyship");
        }
  }
  else {// ss_exploding
	ship->explode_motion ++;
	ship->pos.x += ship->xspeed;
	ship->pos.z += ship->zspeed;
	if (ship->explode_motion == ENEMYSHIP_EXPLODE_MAX){
        /*
	if (ship->explode_motion == ENEMYSHIP_EXPLODE_MAX ||
		ship->pos.x < XMIN + 64 || ship->pos.x > XMAX - 64 ||
		ship->pos.z < ZMIN + 32 || ship->pos.z > ZMAX - 256){
        */
        if (ship->type == et_gunroid){
            // GUNROID撃破時は残っているBYTEをすべて爆破する
            // 撃破ではないのでこの分のBYTEはKILLやスコアにカウントしない
            for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
                if (enemyship[i].state == ss_alive){
                    enemyship[i].state = ss_exploding;
                }
            }
        }
	  delete_enemyship(ship);
	}
  }
}

void move_scout(enemyship_t *ship){
    ship->pos.z += ship->zspeed;
    if (ship->pos.z < ZMIN){
        delete_enemyship(ship);
        return;
  }
  ship->pos.x += ship->xspeed;
  if (ship->pos.x < XMIN || ship->pos.x > XMAX){
	ship->xspeed = -ship->xspeed;
  }
}


void move_wall(enemyship_t *ship){
    ship->pos.z += ship->zspeed;
    if (ship->pos.z < ZMIN){
        delete_enemyship(ship);
        return;
    }
    ship->motion ++;
    if (ship->motion == 12){
        ship->motion = 0;
    }
}

void move_drone(enemyship_t *ship){
    int xd, xl, dir;

    xd = ship->pos.x - myship.pos.x;
    xl = abs(xd);

    if (xl > DRONE_DIST + DRONE_XSPEED){
        if (xd < 0){
            dir = 1;
        }
        else {
            dir = -1;
        }
    }
    else if (xl < DRONE_DIST){
        if (xd < 0){
            dir = -1;
        }
        else {
            dir = 1;
        }
    }
    else {
        dir = 0;
    }
    ship->xspeed = DRONE_XSPEED * dir;
    ship->pos.x += ship->xspeed;
    if (ship->pos.x < XMIN + 32){
        ship->pos.x = XMIN + 32;
    }
    else if (ship->pos.x > XMAX - 32){
        ship->pos.x = XMAX - 32;
    }

    ship->zspeed ++;
    ship->pos.z += ship->zspeed;
    

    if (ship->pos.z < ZMIN || ship->pos.z > ZMAX){
        delete_enemyship(ship);
        return;
    }
    ship->motion ++;
    if (ship->motion == 12){
        ship->motion = 0;
    }
    ship->fire_counter ++;
    if (ship->fire_counter >= ship->fire_rate){
        enemy_fire_aiming(ship->pos, DRONE_BULLET_SPEED, ship->type);
        ship->fire_counter = 0;
        ship->fire_rate = DRONE_FIRE_RATE + RRnd(-6, 6) - stage;
        if (ship->fire_rate < 4){
            ship->fire_rate = 4;
        }
        //        printf("fire rate %d\n", ship->fire_rate);
    }
}
 
void move_fighter(enemyship_t *ship){
    ship->pos.z += ship->zspeed;
    if (ship->pos.z < ZMIN || ship->pos.z > ZMAX){
        delete_enemyship(ship);
        return;
    }
    //    printf("fighter:%d myship:%d\n", ship->pos.z, myship.pos.z);
    if (ship->motion == 11){
        // 帰りは行きより速度を落とす(自機と同方向に進んでいるため)
        ship->zspeed = FIGHTER_ZSPEED/2;
    }
    else if (ship->motion > 0){
        ship->zspeed += FIGHTER_ZSPEED/16;
        ship->pos.y -= 8;
        ship->motion ++;
    }
    else if (ship->pos.z < myship.pos.z + FIGHTER_TURNOFFSET){
        enemy_fire_forward(ship->pos, FIGHTER_BULLET_SPEED, ship->type);
        ship->motion = 1;
        ship->zspeed = 0;
    }
}

void move_copter(enemyship_t *ship){
  ship->pos.x += ship->xspeed;
  ship->pos.z += ship->zspeed;
  if(ship->motion % 3 == 0 && ship->zspeed < 0){
	if (ship->xspeed > 0){
	  ship->xspeed ++;
	}
	else {
	  ship->xspeed --;
	}
	ship->zspeed ++;
  }
  if (ship->pos.x < XMIN || ship->pos.x > XMAX || ship->pos.z < ZMIN){
	delete_enemyship(ship);
    return;
  }
  ship->motion ++;
  if (ship->motion == 6){
	ship->motion = 0;
  }
  ship->fire_counter ++;
  if (ship->fire_counter >= ship->fire_rate){
	enemy_fire_aiming(ship->pos, COPTER_BULLET_SPEED, ship->type);
	ship->fire_counter = 0;
	ship->fire_rate = COPTER_FIRE_RATE + RRnd(-6, 6) - stage;
	if (ship->fire_rate < 4){
	  ship->fire_rate = 4;
	}
	printf("fire rate %d\n", ship->fire_rate);
  }
}

void move_lancer(enemyship_t *ship){
  ship->pos.z += ship->zspeed;
  if (ship->pos.x > myship.pos.x){
	ship->xspeed -= 4;
  }
  else {
	ship->xspeed += 4;
  }
  ship->pos.x += ship->xspeed;
  if (ship->pos.z < ZMIN){
	delete_enemyship(ship);
    return;
  }
}


void move_walker(enemyship_t *ship){
  ship->pos.x += ship->xspeed;
  ship->pos.z += ship->zspeed;
  if (ship->pos.x < XMIN || ship->pos.x > XMAX || ship->pos.z < ZMIN){
	delete_enemyship(ship);
    return;
  }
  ship->motion ++;
  if (ship->motion == 16){
	ship->motion = 0;
  }
  else if (ship->motion == 32){
	ship->motion = 16;
  }
  ship->fire_counter ++;
  if (ship->fire_counter >= ship->fire_rate){
	enemy_fire_aiming(ship->pos, WALKER_BULLET_SPEED, ship->type);
	ship->fire_counter = 0;
	ship->fire_rate = COPTER_FIRE_RATE + RRnd(-6, 6) - stage;
	if (ship->fire_rate < 4){
	  ship->fire_rate = 4;
	}
	printf("fire rate %d\n", ship->fire_rate);
  }
}

void move_turret(enemyship_t *ship){
    ship->pos.z += ship->zspeed;
    if (ship->pos.z < ZMIN){
        delete_enemyship(ship);
        return;
    }
    if (ship->pos.z < TURRET_FIRE_ZPOS){
        ship->motion ++;
        if (ship->motion == 6){
            ship->motion = 0;
        }
        ship->fire_counter ++;
        if (ship->fire_counter >= ship->fire_rate){
            // TURRETは等間隔で前方に射撃
            enemy_fire_forward(ship->pos, TURRET_BULLET_SPEED, ship->type);
            ship->fire_counter = 0;
        }
    }
}

void move_fortress(enemyship_t *ship){
  int i, gun;
  ship->pos.z += ship->zspeed;
  ship->pos.x += ship->xspeed;
  if (ship->pos.z < ZMIN + ZMAX/3 || ship->pos.z > ZMAX - 120){
	ship->zspeed = -ship->zspeed;
  }
  ship->pos.x += ship->xspeed;
  if (ship->pos.x < XMIN + 192 || ship->pos.x > XMAX - 192){
	ship->xspeed = -ship->xspeed;
  }
  ship->motion ++;
  if (ship->motion == 30){
	ship->motion = 0;
  }

  // 画面上方に位置していればSTAGE数分の
  // ねらい撃ち弾を発射
  ship->fire_counter ++;
  if (ship->fire_counter >= ship->fire_rate){
      //	printf("fire rate %d\n", ship->fire_rate);
	if (ship->pos.z > ZMIN + ZMAX/2){
        for (i = 0 ; i < 1 + BossRound() ; i ++){
            gun = RRnd(6, 11);
            enemy_fire_aiming(add_vector3d(ship->pos,
                                           enemyship_model[et_fortress].
                                           vertex[ship->motion][gun]),
                              FORTRESS_BULLET_SPEED, ship->type);
        }
	}
	ship->fire_counter = 0;
	ship->fire_rate = FORTRESS_FIRE_RATE - BossRound();
	if (ship->fire_rate < 4){
	  ship->fire_rate = 4;
	}
  }
}


void move_bomber(enemyship_t *ship){
    //    printf("move_bomber(%d,%d)\n", ship->pos.x, ship->pos.z);
    ship->pos.z += ship->zspeed;
    // 一定以上の距離を飛行したら、BOMBERは消える(撃ち漏らし)
    if (ship->pos.z > ZMAX - 128){
        delete_enemyship(ship);
        return;
    }
    ship->motion ++;
    if (ship->motion < 14){
        return;
    }
    // 飛行
    if (ship->pos.y == FLIGHT_HEIGHT){
        //        if (ship->pos.z > ZMAX-256){
        //            ship->zspeed = -1;
        //        }
        if (ship->pos.z < ZMIN + 256){
            ship->zspeed = 1;
        }
        //        printf("flying\n");
        ship->can_hit = -1;  // 巡航体制に入ったら当たり判定が発生

        ship->fire_counter ++;
        if (ship->fire_counter >= ship->fire_rate){
            //            printf("fire rate %d\n", ship->fire_rate);
            enemy_fire_aiming(ship->pos,
                              BOMBER_BULLET_SPEED, ship->type);
            ship->fire_counter = 0;
            ship->fire_rate = BOMBER_FIRE_RATE - BossRound();
            if (ship->fire_rate < 4){
                ship->fire_rate = 4;
            }
        }
    }
    // 離陸
    else if (ship->zspeed >= 0){
        //        printf("take off\n");
        ship->pos.y -= 2;
        ship->zspeed = 1;
    }
    // 滑走
    else if (ship->pos.z < ZMIN + 140){
        ship->zspeed += 1;
    }
    if (ship->motion == 14+3){
        ship->motion = 14;
    }
}

void move_wyvern(enemyship_t *ship){
    // 登場時は高々度飛行
    if (ship->pos.y < FLIGHT_HEIGHT){
        // 自機の上空を飛び越えて奥に突進
        if (ship->pos.z < ZMAX - 256){
            ship->pos.z += ship->zspeed;
        }
        // あるていど奥に進んだら高度を下げる
        else {
            ship->pos.y += 8;
        }
    }
    // 通常高度まで下がったら戦闘に入る
    else {
        ship->can_hit = -1;  // 当たり判定発生
        ship->pos.x += ship->xspeed;
        ship->pos.z += ship->zspeed;
        if (ship->pos.x < XMIN+64 || ship->pos.x > XMAX-64){
            ship->xspeed = -ship->xspeed;
        }
        if (ship->pos.z < ZMIN+64){
            ship->zspeed = WYVERN_SPEED;
        }
        else if (ship->pos.z > ZMAX-128){
            ship->zspeed = -WYVERN_SPEED;
            // 一番奥に移動した時、その時点での自機の位置に向かって移動を開始する
            // たまたまZ座標が一致した場合は移動ベクトルを変えない
            if (ship->pos.z != myship.pos.z){
                ship->xspeed = ship->zspeed
                    * (ship->pos.x-myship.pos.x) / (ship->pos.z-myship.pos.z);
            }
        }
        // 前進時のみ弾を発射する
        // 後退時は自機への突進なので弾は撃たない
        if (ship->zspeed > 0){
            ship->fire_counter ++;
            if (ship->fire_counter >= ship->fire_rate){
                printf("fire rate %d\n", ship->fire_rate);
                enemy_fire_spread(ship->pos,
                                  WYVERN_BULLET_SPEED, 2 + BossRound(),
                                  ship->type);
                ship->fire_counter = 0;
                ship->fire_rate = WYVERN_FIRE_RATE - BossRound();
                if (ship->fire_rate < 4){
                    ship->fire_rate = 4;
                }
            }
        }
    }
    if (ship->motion  >= 8 && ship->motion < 16){
        ship->pos.y -= 4;
    }
    else {
        ship->pos.y += 2;
    }
    ship->motion ++;
    if (ship->motion >= 24){
        ship->motion = 0;
    }

}

void move_byte(enemyship_t *ship, int no){
    int dx, dz;
    if (current_group.type == et_byte){
        switch (ship->counter){
        case 0:
            // 地上に置かれたまま近づいてくる
            if (ship->pos.z > 480){
                ship->pos.z += ship->zspeed;
            }
            else {
                ship->counter = 1;
            }
            break;
        case 1:
            // 浮上
            if (ship->pos.y > FLIGHT_HEIGHT){
                ship->pos.y += (FLIGHT_HEIGHT - GROUND_HEIGHT)/8;
            }
            else {
                ship->can_hit = -1;  // 当たり判定発生
                ship->counter = 2;
            }
            break;
        case 2:
            dx = RRnd(XMIN, XMAX);
            dz = RRnd(ZMIN+240, ZMAX-160);
            ship->xspeed = (dx - ship->pos.x)/BYTE_MOVESTEP;
            ship->zspeed = (dz - ship->pos.z)/BYTE_MOVESTEP;
            ship->counter = 3;
        case 3:
            ship->pos.x += ship->xspeed;
            ship->pos.z += ship->zspeed;
            ship->fire_counter++;
            if (ship->fire_counter == BYTE_MOVESTEP){
                ship->counter = 4;
                ship->fire_counter = 0;
            }
            break;
        case 4:
            ship->fire_counter ++;
            if (ship->fire_counter >= BYTE_CHARGE_TIME){
                enemy_fire_forward(ship->pos, BYTE_BULLET_SPEED, ship->type);
                ship->fire_counter = 0;
                ship->counter = 2;
            }
        }
    }
    // GUNROIDの制御下
    else {
        //        printf("byte counter:%d\n", ship->counter);
        switch (ship->counter){
        case 0:
            // アサルトモード開始
            if (IsAssaultL()) {
                // モードL
                dx = (no - (BYTE_NUM+1)/2) * 100;
                dz = BYTE_ASSAULT_ZPOS;
            }
            else {
                // モードR
                dx = RRnd(XMIN, XMAX);
                dz = RRnd(ZMIN+240, ZMAX-160);
            }
            ship->xspeed = (dx - ship->pos.x)/BYTE_MOVESTEP;
            ship->zspeed = (dz - ship->pos.z)/BYTE_MOVESTEP;
            ship->counter = 1;
            ship->fire_counter = 0;
            break;
        case 1:
            // アサルトモード移動
            ship->pos.x += ship->xspeed;
            ship->pos.z += ship->zspeed;
            ship->fire_counter ++;
            if (ship->fire_counter == BYTE_MOVESTEP){
                ship->counter = 2;
            }
            break;
        case 2:
            // アサルトモード射撃
            if (ship->motion < 8){
                ship->motion ++;
            }
            else {
                enemy_fire_forward(ship->pos, BYTE_BULLET_SPEED, ship->type);
                ship->counter = 3;
            }
            break;
        case 3:
            if (ship->motion > 0){
                ship->motion --;
            }
            else {
                ship->counter = 4;
            }
            break;
        case 4:
            // シールドモード開始
            dx = (no - (BYTE_NUM+1)/2) * 40;
            dz = BYTE_SHIELD_ZPOS;
            ship->xspeed = (dx - ship->pos.x)/BYTE_MOVESTEP;
            ship->zspeed = (dz - ship->pos.z)/BYTE_MOVESTEP;
            ship->counter = 5;
            ship->fire_counter = 0;
            break;
        case 5:
            // シールドモード移動
            ship->pos.x += ship->xspeed;
            ship->pos.z += ship->zspeed;
            ship->fire_counter ++;
            if (ship->fire_counter == BYTE_MOVESTEP){
                ship->counter = 6;
                ship->fire_counter = 0;
            }
            break;
        case 6:
            // シールドモードで一定時間防御
            if (ship->fire_counter < BYTE_SHIELD_TIME){
                ship->fire_counter ++;
            }
            else {
                ship->counter = 0;
            }
            break;
        }
    }
}


void move_gunroid(enemyship_t *ship){
    int dx, dz;
    int no;
    enemyship_t *byte;

    //    printf("GUNROID tracing BYTE %d %d\n", ship->fire_counter, ship->pos.z);
    switch (ship->counter){
    case 0:
        if (ship->pos.y > FLIGHT_HEIGHT){
            // 一定距離まで前進
            if (ship->pos.z > BYTE_SHIELD_ZPOS + 140){
                ship->pos.z -= GRIDSIZE+16;
            }
            // その後、飛行高度まで上昇
            else {
                ship->pos.z -= 8;
                ship->pos.y -= 8;
            }
        }
        // 飛行高度に達したらBYTEを展開して戦闘態勢に入る
        else {
            printf("create byte\n");
            no = allocate_enemyship();
            if (no != -1){
                byte =  &enemyship[no];
                byte->type = et_byte;
                create_byte(byte);
                ship->counter = 1;
            }
            ship->fire_counter = BYTE_NUM/2+1;
            ship->can_hit = -1;  // 当たり判定発生
            ship->counter = 1;
        }
        break;
    case 1:
        byte = &enemyship[ship->fire_counter];
        if (byte->state != ss_alive){
            ship->fire_counter = find_alive_enemyship(et_byte);
            if (ship->fire_counter == -1){
                ship->counter = 2;
                return;
            }
            byte = &enemyship[ship->fire_counter];
        }
        switch (byte->counter){
        case 0:
            // アサルトモード開始
            dx = RRnd(XMIN, XMAX);
            dz = RRnd(ZMIN+240, ZMAX-160);
            ship->xspeed = (dx - ship->pos.x)/BYTE_MOVESTEP;
            ship->zspeed = (dz - ship->pos.z)/BYTE_MOVESTEP;
            break;
        case 1:
            // アサルトモード移動
            ship->pos.x += ship->xspeed;
            ship->pos.z += ship->zspeed;
            break;
        case 2:
            if (byte->motion == 8){
                // BYTEが半減していたら拡散弾を撃つ
                if (enemyship_num <= BYTE_NUM/2){
                    enemy_fire_spread(ship->pos, BYTE_BULLET_SPEED, 2+BossRound(),
                                      ship->type);
                }
                else {
                    enemy_fire_forward(ship->pos, BYTE_BULLET_SPEED, ship->type);
                }
            }
            break;
        case 3:
            break;
        case 4:
            // シールドモードに向けて移動
            dx = (ship->fire_counter - (BYTE_NUM+1)/2) * 40;
            dz = GUNROID_DEFENCE_ZPOS;
            ship->xspeed = (dx - ship->pos.x)/BYTE_MOVESTEP;
            ship->zspeed = (dz - ship->pos.z)/BYTE_MOVESTEP;
            break;
        case 5:
            // シールドモード移動
            ship->pos.x += ship->xspeed;
            ship->pos.z += ship->zspeed;
        case 6:
            // シールドモードで一定時間防御
            break;
        }
        
        break;
    case 2:
        // BYTEが残っていなければ、独自の動きをする
        dx = RRnd(XMIN, XMAX);
        dz = RRnd(ZMIN+80, ZMAX-160);
        ship->xspeed = (dx - ship->pos.x)/BYTE_MOVESTEP;
        ship->zspeed = (dz - ship->pos.z)/BYTE_MOVESTEP;
        ship->counter = 3;
        ship->fire_counter = 0;
        break;
    case 3:
        ship->pos.x += ship->xspeed;
        ship->pos.z += ship->zspeed;
        ship->fire_counter ++;
        if (ship->fire_counter == BYTE_MOVESTEP){
            ship->counter = 4;
            ship->fire_counter = 0;
        }
        break;
    case 4:
        // 拡散弾を撃つ
        ship->fire_counter ++;
        if (ship->fire_counter == BYTE_MOVESTEP){
            enemy_fire_spread(ship->pos, BYTE_BULLET_SPEED, 2 + BossRound(), ship->type);
            ship->counter = 2;
        }
        break;
    }
}

void move_dragon_head(enemyship_t *ship){
    int dx, dz, xl, zl, step;

    assert(dragon != NULL);
    switch (ship->action){
    case ea_dragon_head_set_dest:
        dx = RRnd(dragon->pos.x - DRAGON_HEAD_MOVE_RANGE, 
                 dragon->pos.x + DRAGON_HEAD_MOVE_RANGE);
        dz = RRnd(dragon->pos.z - DRAGON_HEAD_MOVE_RANGE, 
                 dragon->pos.z + DRAGON_HEAD_MOVE_RANGE);
        xl = abs(dx - ship->pos.x);
        zl = abs(dz - ship->pos.z);
        if (xl > zl){
            step = xl / DRAGON_HEAD_SPEED;
        }
        else {
            step = zl / DRAGON_HEAD_SPEED;
        }
        if (step == 0) step = 1;
        ship->xspeed = (dx - ship->pos.x)/step;
        ship->zspeed = (dz - ship->pos.z)/step;
        ship->counter = step;
        ship->action = ea_dragon_head_move;
        break;
    case ea_dragon_head_move:
        ship->counter --;
        if (ship->counter == 0){
            // 登場時の本体前進中は、首を動かすだけでブレスを吐かない
            if (dragon->action == ea_dragon_engage){
                ship->action = ea_dragon_head_set_dest;
            }
            else {
                ship->action = ea_dragon_head_open_mouth;
            }
        }
        break;
    case ea_dragon_head_open_mouth:
        ship->xspeed = dragon->xspeed;
        ship->zspeed = dragon->zspeed;
        ship->motion ++;
        if (ship->motion % 6 == 5){
            ship->action = ea_dragon_head_breath;
            ship->fire_counter = BREATH_LEN;
        }
        break;
    case ea_dragon_head_breath:
        ship->xspeed = dragon->xspeed;
        ship->zspeed = dragon->zspeed;
        ship->fire_counter --;
        if (ship->fire_counter == 0){
            ship->action = ea_dragon_head_close_mouth;
        }
        break;
    case ea_dragon_head_close_mouth:
        ship->xspeed = dragon->xspeed;
        ship->zspeed = dragon->zspeed;
        ship->motion --;
        if (ship->motion % 6 == 0){
            ship->action = ea_dragon_head_set_dest;        
            if (ship->motion == 0){
                ship->motion += 6;
            }
            else if (ship->motion == 6*4){
                ship->motion -= 6;
            }
            else {
                if (RRnd(0, 1) == 0){
                    ship->motion += 6;
                }
                else {
                    ship->motion -= 6;
                }
            }
        }
        break;
    default:
        assert(0);
        break;
    }
    ship->pos.x += ship->xspeed;
    ship->pos.z += ship->zspeed;
    //    printf("%d h(%d, %d) d(%d, %d) dx:dz %d, %d, spd(%+d, %+d)\n", ship->action, ship->pos.x, ship->pos.z, dragon->pos.x, dragon->pos.z, dx, dz, ship->xspeed, ship->zspeed);
}

#ifdef DUMMY

        ship->fire_counter = step;
        break;
    case 1:
        ship->pos.x += ship->xspeed;
        ship->pos.z += ship->zspeed;
        ship->fire_counter--;
        // 登場時に本体が前進している間は首を動かすだけでブレスを吐かない
        if (dragon->counter == 0){
            ship->counter = 0;
        }
        else {
            if (ship->fire_counter == 0){
                ship->counter = 2;
            }
        }
        break;
    case 2:
        ship->motion ++;
        if (ship->motion % 6 == 5){
            ship->counter = 3;
            ship->fire_counter = 0;
        }
        break;
    case 3:
        ship->fire_counter++;
        if (ship->fire_counter == BREATH_LEN){
            ship->counter = 4;
        }
        break;
    case 4:
        ship->motion --;
        if (ship->motion % 6 == 0){
            ship->counter = 0;
            ship->fire_counter = 0;
            if (ship->motion == 0){
                ship->motion += 6;
            }
            else if (ship->motion == 6*4){
                ship->motion -= 6;
            }
            else {
                if (RRnd(0, 1) == 0){
                    ship->motion += 6;
                }
                else {
                    ship->motion -= 6;
                }
            }
        }
        break;
    }
}
#endif


void move_dragon(enemyship_t *ship){
    int i;

    ship->motion ++;

    if (ship->motion == 24) ship->motion = 0;
    switch (ship->action){
    case ea_dragon_engage:
        if (ship->pos.z < DRAGON_ZPOS){
            ship->zspeed = DRAGON_SPEED;
            ship->xspeed = DRAGON_SPEED;
            ship->action = ea_dragon_wander;
            for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
                if (enemyship[i].type == et_dragon_head){
                    enemyship[i].can_hit = -1;
                }
            }
        } 
        else {
            ship->pos.z += ship->zspeed;
        }
        break;
    case ea_dragon_wander:
        ship->pos.x += ship->xspeed;
        ship->pos.z += ship->zspeed;
        if (ship->pos.x < -120 || ship->pos.x > 120){
            ship->xspeed = -ship->xspeed;
        }
        if (ship->pos.z < DRAGON_ZPOS-160 || ship->pos.z > DRAGON_ZPOS){
            ship->zspeed = -ship->zspeed;
        }
        break;
    default:
        assert(0);
        break;
    }
}

void motion_rattle_core(enemyship_t *ship){
    // 
    // 
    //     6 5 4 0 3 2 1
    //      ---- 1 --->
    //      <--- 2 ----
    if (ship->action == ea_core_yaw_left){
        if (ship->motion < 4){
            ship->motion ++;
            if (ship->motion == 3){
                ship->action = ea_core_yaw_right;
                enemy_fire_aiming(ship->pos, CORE_BULLET_SPEED, et_core);
            }
        }
        else {
            ship->motion --;
            if (ship->motion == 3){
                ship->motion = 0;
            }
        }
    }
    else if (ship->action == ea_core_yaw_right){
        if (ship->motion < 4){
            ship->motion --;
            if (ship->motion == -1){
                ship->motion = 4;
            }
        }
        else {
            ship->motion ++;
            if (ship->motion == 6){
                ship->action = ea_core_yaw_left;
                enemy_fire_aiming(ship->pos, CORE_BULLET_SPEED, et_core);
            }
        }
    }
}

void move_core(enemyship_t *ship){

    switch (ship->action){
    // 出現時は地上なので浮上する
    case ea_core_takeoff:
        ship->pos.y -= 8;
        if (ship->pos.y < FLIGHT_HEIGHT){
            ship->can_hit = -1;  // 当たり判定発生
            ship->zspeed = CORE_SPEED;
            ship->xspeed = CORE_SPEED;
            ship->action = ea_core_yaw_left;
        }
        break;
    case ea_core_yaw_left:
    case ea_core_yaw_right:
        if (ship->counter % 2 == 0) {
            motion_rattle_core(ship);
        }
        ship->pos.x += ship->xspeed;
        ship->pos.z += ship->zspeed;
        if (ship->pos.x < XMIN + 128 || ship->pos.x > XMAX - 128){
            ship->xspeed = -ship->xspeed;
        }
        if (ship->pos.z < DRAGON_ZPOS-160 || ship->pos.z > DRAGON_ZPOS){
            //        if (ship->pos.z < ZMIN - 128 || ship->pos.z > ZMAX - 128){
            ship->zspeed = -ship->zspeed;
        }
        ship->counter ++;
        if (ship->counter > CORE_WANDER_COUNT && ship->motion == 0){
            ship->motion = 8;
            ship->action = ea_core_setup_charge;
        }
        break;
    case ea_core_setup_charge:
        ship->motion ++;
        if (ship->motion == 12){
            ship->motion = 20;
            ship->zspeed = -CORE_SPEED;
            // 一番奥に移動した時、その時点での自機の位置に向かって
            // 移動を開始する
            // たまたまZ座標が一致した場合は移動ベクトルを変えない
            if (ship->pos.z != myship.pos.z){
                ship->xspeed = ship->zspeed
                    * (ship->pos.x-myship.pos.x) / (ship->pos.z-myship.pos.z);
            }
            // z方向が近すぎるときxspeedが巨大になりうるので
            // 横方向のスピードにリミットをかけておく
            if (ship->xspeed > CORE_SPEED) ship->xspeed = CORE_SPEED;
            if (ship->xspeed < -CORE_SPEED) ship->xspeed = -CORE_SPEED;
            ship->action = ea_core_charge;
        }
        break;
    // 画面手前に向かって加速
    case ea_core_charge:
        // 自機が横や後ろにいると突撃方向がおかしいので
        // とりあえず画面両端を超えそうなときは端で跳ね返えらせておく
        if (ship->pos.x < XMIN + 128 || ship->pos.x > XMAX - 128){
            ship->xspeed = -ship->xspeed;
        }
        if (ship->zspeed > -CORE_CHARGE_SPEED){
            ship->zspeed -= 2;
        }
        ship->pos.x += ship->xspeed;
        ship->pos.z += ship->zspeed;
        // 画面手前に近づいたら減速・後退モードに移行
        if (ship->pos.z < ZMIN+128){
            ship->action = ea_core_retreat;
        }
        ship->motion ++;
        if (ship->motion == 32){
            ship->motion = 20;
        }
        break;
    // 向かって画面手前への突撃を減速し画面奥に戻る
    case ea_core_retreat:
        // 横方向の移動量は突撃時のままだが、
        // 画面両端を超えそうなときは端で跳ね返えらせる
        if (ship->pos.x < XMIN + 128 || ship->pos.x > XMAX - 128){
            ship->xspeed = -ship->xspeed;
        }
        // ある程度後退したら、停止に向けて減速開始
        if (ship->pos.z > ZMAX-560){
            if (ship->zspeed > 0){
                ship->zspeed -= 2;
            }
            else {
                ship->xspeed = 0;
            }
        }
        else {
            // 突撃速度に達するまでは後退方向に加速
            if (ship->zspeed < CORE_CHARGE_SPEED){
                ship->zspeed += 2;
            }
        }
        ship->pos.x += ship->xspeed;
        ship->pos.z += ship->zspeed;
        ship->motion ++;
        if (ship->motion == 32){
            
            if (ship->pos.z > ZMIN+128 && ship->zspeed == 0) {
                ship->motion = 12;
                ship->action = ea_core_open_mouth;
                break;
            }
            else {
             ship->motion = 20;
            }
        }
        break;
    case ea_core_open_mouth:
        ship->motion ++;
        if (ship->motion == 19){
            ship->action = ea_core_laser;
            ship->counter = 0;
            // 口をあけ終えた直後に一回目のレーザを発射
            ship->fire_counter = CORE_FIRE_LASER_RATE;
        }
        break;
    case ea_core_laser:
        ship->fire_counter ++;
        if (ship->fire_counter >= CORE_FIRE_LASER_RATE){
            create_lasers(ship);
            ship->fire_counter = 0;
            ship->counter ++;
            if (ship->counter == (stage-1)/3 + 1){
                ship->action = ea_core_close_mouth;
            }
        }
        break;
    case ea_core_close_mouth:
        ship->motion --;
        if (ship->motion == 11){
            ship->motion = 0;
            ship->counter = 0;
            ship->zspeed = CORE_SPEED;
            ship->xspeed = CORE_SPEED;
            ship->action = ea_core_yaw_left;
        }
        break;
    default:
        assert(0);
        break;
    }
    /*
    printf("core (%d,%d) (%+d,%+d) action:%d counter:%d motion:%d\n",
           ship->pos.x, ship->pos.z,
           ship->xspeed, ship->zspeed,
           ship->action, ship->counter, ship->motion);
    */
}


void move_enemybullet(bullet_t *bullet){
  bullet->pos.x += bullet->xa;
  bullet->pos.z += bullet->za;
  if (bullet->pos.x < XMIN || bullet->pos.x > XMAX ||
	  bullet->pos.z < ZMIN || bullet->pos.z > ZMAX){
	bullet->state = bs_dead;
  }
}

void game_keyh(int key){
    switch (key){
    case EZ_Left:
    case EZ_Up:
    case EZ_Right:
    case EZ_Down:
        // キーリピートで同じキーの割り込みが複数回発生するので
        // +=ではなくORをとる必要がある
        dir_keymap |= 1 << (key-EZ_Left);
        dir_prev = dir_table[dir_keymap];
        break;
    case ' ':
        is_fire = 1;
        break;
    case 'p':
        if (gamestate == gs_playing){
            if (is_pause){
                is_pause = 0;
            }
            else {
                is_pause = 1;
            }
        }
        break;
    case 'q' :
        if (gamestate == gs_playing){
            if (myship.state == ss_dead){
                init_title();
            }
            else if (myship.state == ss_alive){
                myship.state = ss_exploding;
                myship.explode_motion = 0;
            }
        }
        break;
    case 'v':
        toggle_view_mode();
        break;
    default:
        if (key == boss_pass[boss_ix]){
            boss_ix ++;
            //            printf("BOSS:%c %d %c '%s'\n", key, boss_ix, boss_pass[boss_ix],
            //                   boss_pass);
            if (boss_pass[boss_ix] == '\0'){
                switch (config.options[ci_stage]){
                case gm_original:
                    group_no = group_num - 2;
                    break;
                case gm_extended:
                    if (group_no < midboss_index){
                        group_no = midboss_index-1;
                    }
                    else {
                        group_no = group_num - 2;
                    }
                    break;
                case gm_boss:
                    break;
                }
                printf("BOSS password '%s' accepted. jump to group %d\n",
                       boss_pass, group_no);
                current_group.enemy_num = 0;
                boss_ix = 0;
                is_cheat = 1;
            }
        }
    }
}

void game_key_releaseh(int key){
    switch (key){
    case EZ_Left:
    case EZ_Up:
    case EZ_Right:
    case EZ_Down:
		// key_release割り込みはリピートしないが、
		// このハンドラが登録されていないところでキーが押されたときはdir_keymapに
        // 加算されないので、単純に引き算すると引きすぎる可能性がある
        // このためANDでビットを潰す
        dir_keymap &= (0xf - (1 << (key-EZ_Left)));
        break;
    case ' ':
        is_fire = 0;
        break;
    }
}


void fire(void){
  int i;
  for (i = 0 ; i < MYBULLET_MAX ; i++){
	if (mybullet[i].state == bs_dead){
	  mybullet[i].state = bs_alive;
	  mybullet[i].pos = myship.pos;
      break;
	}
  }
}

void enemy_fire_forward(point3d_t pos, int speed, enemytype_t type){
    int i;

    printf("aiming forward!\n");
    i = enemy_create_bullet(pos, type);
    if (i > -1){
        enemybullet[i].xa = 0;
        enemybullet[i].za = -speed;
    }
}


void enemy_fire_aiming(point3d_t pos, int speed, enemytype_t type){
    int i, l, lx, lz;

    printf("aiming fire!\n");
    i = enemy_create_bullet(pos, type);
    if (i > -1){
        lx = myship.pos.x - pos.x;
        lz = myship.pos.z - pos.z;
        l = (int)sqrt(lx*lx + lz*lz);
        enemybullet[i].xa = lx * speed / l;
        enemybullet[i].za = lz * speed / l;
    }
}


// 自機の方向へ45度単位で狙いをつけ、num*2-1個の弾を撃つ
// 速度ベクトルの大きさがspeedになるようにしていないので、
// 斜め方向の弾は速度が1.4倍になっている
void enemy_fire_spread(point3d_t pos, int speed, int num, enemytype_t type){
    int i, j, lx, lz, xa, za, xs, zs;

    printf("spread fire!\n");
    lx = myship.pos.x - pos.x;
    lz = myship.pos.z - pos.z;

    if (lx == 0 || abs(lz/lx) > 2){
        printf("fire up/down\n");
        xa = 0;
        za = (lz < 0) ? -1 : 1;
        xs = 1;
        zs = 0;
    }
    else if (lz == 0 || abs(lx/lz) > 2){
        printf("fire left/right\n");
        xa = (lx < 0) ? -1 : 1;
        za = 0;
        xs = 0;
        zs = 1;
    }
    else {
        printf("fire diagonal\n");
        xa = (lx < 0) ? -1 : 1;
        za = (lz < 0) ? -1 : 1;
        xs = xa;
        zs = -za;
    }

    i = enemy_create_bullet(pos, type);
    if (i > -1){
        enemybullet[i].xa = xa * speed;
        enemybullet[i].za = za * speed;
    }
    for (j = -num+1 ; j < num ; j++){
        if (j != 0){
            i = enemy_create_bullet(pos, type);
            if (i > -1){
                enemybullet[i].xa = xa * speed + xs * 20*j;
                enemybullet[i].za = za * speed + zs * 20*j;
            }
        }
    }
}


int enemy_create_bullet(point3d_t pos, enemytype_t type){
    int i;
    for (i = 0 ; i < ENEMYBULLET_MAX ; i++){
        if (enemybullet[i].state == bs_dead){
            enemybullet[i].state = bs_alive;
            enemybullet[i].pos = pos;
            enemybullet[i].owner_type = type;
            return i;
        }
    }
    return -1;
}

void create_lasers(enemyship_t *ship){
    int i, laser_num;
    static int xoffsets[] = {  8, 16, 24, 16};
    static int zoffsets[] = {-10, -6, -2,  2};
    
    laser_num = ((stage-1)/3 + 1) * 2;
    if (laser_num > 8) laser_num = 8;
    
    for (i = 0 ; i < laser_num ; i++){
        if (i % 2 == 0){
            create_laser(ship, -xoffsets[i/2], zoffsets[i/2]);
        }
        else {
            create_laser(ship, xoffsets[i/2], zoffsets[i/2]);
        }
    }
    
}
void create_laser(enemyship_t *ship, int vx0, int vz0){
    int i;
    laser_t *laser;
    for (i = 0 ; i < LASER_MAX_NUM ; i++){
        if (lasers[i].is_valid == 0){
            laser = &lasers[i];
            laser->is_valid = 1;
            laser->head = 0;
            laser->tail = 1;
            laser->nodes[0].t = ship->pos;
            laser->vx = vx0;
            laser->vz = vz0;
            laser->nodes[1].t.x = laser->nodes[0].t.x + laser->vx;
            laser->nodes[1].t.y = laser->nodes[0].t.y;
            laser->nodes[1].t.z = laser->nodes[0].t.z + laser->vz;
            printf("laser created (%d, %d, %d)(%d, %d, %d)\n",
                   laser->nodes[0].t.x,
                   laser->nodes[0].t.y,
                   laser->nodes[0].t.z,
                   laser->nodes[1].t.x,
                   laser->nodes[1].t.y,
                   laser->nodes[1].t.z
                   );
            break;
        }
    }
}


void init(int argc, char *argv[]){
    int i;

    if (getarg(argc, argv) != 0){
        exit(1);
    }
    file_init(argv[0], get_opt_str("savedir"), SAVE_DIR, LOG_FILE);
    if (!load_config()){
        log_warn("Loading '%s/%s' failed.\n",
                 get_save_dpath(), CONFIG_FILE);
    }
    if (!load_achievements()){
        log_warn("Loading '%s/%s' failed.\n",
                 get_save_dpath(), ACHIEVEMENTS_FILE);
    }
    if (achievements.top_stage[gm_original] <= 1 &&
        config.options[ci_stage] > 0)
    {
        // ロックされたモードを選択していたらモードを強制的にORIGINALにする
        config.options[ci_stage] = 0;
    }
    else if (achievements.top_stage[gm_extended] <= 3 &&
             config.options[ci_stage] > 1)
    {
        // ロックされたモードを選択していたらモードを強制的にEXTENDEDにする
        config.options[ci_stage] = 1;
    }
    set_view(config.options[ci_view]);
    set_color_set_type(config.options[ci_colorset]);

    srand(1);
    enemyship_model[et_scout] = scout_model;
    enemyship_model[et_wall] = wall_model;
    enemyship_model[et_drone] = drone_model;
    enemyship_model[et_fighter] = fighter_model;
    enemyship_model[et_copter] = copter_model;
    enemyship_model[et_lancer] = lancer_model;
    enemyship_model[et_walker] = walker_model;
    enemyship_model[et_turret] = turret_model;
    enemyship_model[et_fortress] = fortress_model;
    enemyship_model[et_bomber] = bomber_model;
    enemyship_model[et_wyvern] = wyvern_model;
    enemyship_model[et_byte] = byte_model;
    enemyship_model[et_gunroid] = gunroid_model;
    enemyship_model[et_dragon_head] = dragon_head_model;
    enemyship_model[et_dragon] = dragon_model;
    enemyship_model[et_core] = core_model;

    create_motion_myship(&myship_model);
    // 自機の姿勢毎の爆発パターンを作るため、一時的に
    // motion_enを設定
    //    myship_model.motion_en = 2;
    create_explosion(&myship_model);
    //    myship_model.motion_en = 0;
    boss_pass[5] = (1 << 6) + (1 << 5) + (1 << 4) - (1<<3) - (1 << 2) + 1;

    //  create_motion_scout(&enemyship_model[et_scout]);
    create_motion_wall(&enemyship_model[et_wall]);
    create_motion_drone(&enemyship_model[et_drone]);
    create_motion_fighter(&enemyship_model[et_fighter]);
    create_motion_copter(&enemyship_model[et_copter]);
    //  create_motion_lancer(&enemyship_model[et_lancer]);
    create_motion_walker(&enemyship_model[et_walker]);
    create_motion_turret(&enemyship_model[et_turret]);
    create_motion_fortress(&enemyship_model[et_fortress]);
    create_motion_bomber(&enemyship_model[et_bomber]);
    create_motion_wyvern(&enemyship_model[et_wyvern]);
    create_motion_byte(&enemyship_model[et_byte]);
    create_motion_gunroid(&enemyship_model[et_gunroid]);
    create_motion_dragon_head(&enemyship_model[et_dragon_head]);
    create_motion_dragon(&enemyship_model[et_dragon]);
    create_motion_core(&enemyship_model[et_core]);
    for (i = et_scout ; i < et_end ; i++){
        create_explosion(&enemyship_model[i]);
    }
    for (i = 0 ; i < mode_group_num[gm_extended] ; i++){
        if (group_extended[i].type == et_midboss){
            extended_midboss_index = i;
            //            printf("midboss found:%d\n", i);
        }
        else if (group_extended[i].type == et_boss){
            //            printf("boss found:%d\n", i);
            extended_boss_index = i;
        }
    }
    for (i = 0 ; i < mode_group_num[gm_boss] ; i++){
        if (group_boss[i].type == et_midboss){
            boss_midboss_index = i;
            //            printf("midboss found:%d\n", i);
        }
        else if (group_boss[i].type == et_boss){
            //            printf("boss found:%d\n", i);
            boss_boss_index = i;
        }
    }
    create_motion_breath();

    EzOpenWindow(WIN_WIDTH, WIN_HEIGHT);
    // init内でsrand固定で乱数を使うので、ここで無作為系列に設定
    srand((unsigned int)time(NULL));
#if defined(FONT)
    EzSetFont(FONT);
#endif
    EzSetFontSize(FONT_SIZE);
    init_title();
}

void init_title(void){
  gamestate = gs_title;
  gamestate_counter = 0;
  myship.state = ss_dead;
  EzSetKeyHandler(title_keyh);
  EzSetTimerHandler(title_timerh, GAME_SPEED);
}

void init_achievements(void){
  gamestate = gs_achievements;
  gamestate_counter = 0;
  EzSetKeyHandler(achievements_keyh);
  EzSetTimerHandler(achievements_timerh, GAME_SPEED);
}

void init_config(void){
  gamestate = gs_config;
  gamestate_counter = 0;
  config_item = ci_stage;
  // Original, Extendedを一周するまでは、Extended, Bossモードをロック
  if (achievements.top_stage[gm_original] <= 1){
      options_num[ci_stage] = 1;
  }
  else if (achievements.top_stage[gm_extended] <= 3){
      options_num[ci_stage] = 2;
  }
  else {
      options_num[ci_stage] = 3;
  }
  EzSetKeyHandler(config_keyh);
  EzSetTimerHandler(config_timerh, GAME_SPEED);
}

void init_game(void){
    int i;

    gamestate = gs_playing;

    // 通常のプレイならプレイ回数加算
    if (!is_cheat){
        IncUInt(achievements.total_play);
    }
    init_myship();

    for (i = 0 ; i < MYBULLET_MAX ; i++){
        mybullet[i].state = bs_dead;
    }
    enemyship_num = 0;
    for (i = 0 ; i < ENEMYSHIP_MAX ; i++){
        enemyship[i].state = ss_dead;
    }
    for (i = 0 ; i < ENEMYBULLET_MAX ; i++){
        enemybullet[i].state = bs_dead;
    }
    score = 0;
    myship.left = config.options[ci_myship];
    group_no = 0;

    group = mode_group[config.options[ci_stage]];
    group_num = mode_group_num[config.options[ci_stage]];
    if (config.options[ci_stage] == gm_extended){
        midboss_index = extended_midboss_index;
        boss_index = extended_boss_index;
    }
    else if (config.options[ci_stage] == gm_boss){
        midboss_index = boss_midboss_index;
        boss_index = boss_boss_index;
    }
    game_speed = GAME_SPEED;
    boss_ix = 0;
    boss_pass[4] = 0xff - 0x9a;

    printf("group_num %d %d\n", group_num, group_no);

    // 初回起動時は初回プレイ開始時にtop_stageを1にする
    set_stage(1);
    // set_stage()で中ボス・ボスをgroupに設定するので
    // current_groupはset_stage()後に設定する必要がある。
    current_group = group[group_no];

    is_autofire = config.options[ci_autofire];

    EzSetKeyHandler(game_keyh);
    EzSetKeyReleaseHandler(game_key_releaseh);
    EzSetTimerHandler(game_timerh, game_speed);
}


void init_myship(void){
  myship.state = ss_alive;
  myship.pos.x = 0;
  myship.pos.z = ZMIN + 128;
  myship.xa = 0;
  myship.za = 0;
}

void init_boss_life(int base_life, int add_life){
    if (config.options[ci_stage] == gm_original){
        boss_life = base_life + add_life * (stage-1);
    }
    // ORIGINALモード以外は3周しないと同じボスが出ないので、
    // ボスの体力は3周ごとに増やす
    else {
        boss_life = base_life + (int)((stage-1) / 3) * add_life;
    }
    boss_life_max = boss_life;
}

void quit(void){
    if (!save_achievements()){
        log_error("saving '%s/%s' failed.\n",
                  get_save_dpath(), ACHIEVEMENTS_FILE);
    }
    if (!save_config()){
        log_error("saving '%s/%s' failed.\n",
                  get_save_dpath(), CONFIG_FILE);
    }
    printf("Bye.\n");
}

int load_achievements(void){
    FILE *fp;
    unsigned char file_major_version = 0, file_minor_version = 0;

    if ((fp = open_data(ACHIEVEMENTS_FILE, "rb",
                        APPLICATION_ID, APPLICATION_ID_SIZE,
                        &file_major_version, &file_minor_version)) == NULL)
    {
        return 0;
    }
    // 旧バージョンなら読み込み中止
    if (GetVersion(file_major_version, file_minor_version) <
        GetVersion(1, 70))
    {
        log_warn("Cannot read old achievement file (Ver%d.%2d)... renamed.\n",
                 file_major_version, file_minor_version);
        fclose(fp);
        rename_data(ACHIEVEMENTS_FILE);
        return 0;
    }
    fread_uint(fp, achievements.top_score, gm_reserve,
               ACHIEVEMENTS_FILE, "top_score");
    fread_uint(fp, achievements.top_stage, gm_reserve,
               ACHIEVEMENTS_FILE, "top_stage");
    fread_uint(fp, &achievements.total_play, 1,
               ACHIEVEMENTS_FILE, "total_play");
    fread_uint(fp, achievements.total_kill_type, et_reserve,
               ACHIEVEMENTS_FILE, "top_kill_type");
    fread_uint(fp, achievements.total_crush_type, et_reserve,
               ACHIEVEMENTS_FILE, "total_crush_type");
    fread_uint(fp, achievements.total_shotdown_type, et_reserve,
               ACHIEVEMENTS_FILE, "total_shotdown_type");
    fclose(fp);
    return 1;
}

int save_achievements(void){
    FILE *fp;

    if ((fp = open_data(ACHIEVEMENTS_FILE, "wb",
                        APPLICATION_ID, APPLICATION_ID_SIZE,
                        &major_version, &minor_version)) == NULL)
    {
        return 0;
    }
    fwrite_uint(fp, achievements.top_score, gm_reserve);
    fwrite_uint(fp, achievements.top_stage, gm_reserve);
    fwrite_uint(fp, &achievements.total_play, 1);
    fwrite_uint(fp, achievements.total_kill_type, et_reserve);
    fwrite_uint(fp, achievements.total_crush_type, et_reserve);
    fwrite_uint(fp, achievements.total_shotdown_type, et_reserve);
    fclose(fp);
    return 1;
}

int load_config(void){
    FILE *fp;
    unsigned char file_major_version = 0, file_minor_version = 0;

    if ((fp = open_data(CONFIG_FILE, "rb", APPLICATION_ID, APPLICATION_ID_SIZE,
                        &file_major_version, &file_minor_version)) == NULL)
    {
        return 0;
    }
    // 旧バージョンなら読み込み中止
    if (GetVersion(file_major_version, file_minor_version) <
        GetVersion(2, 00))
        //        GetVersion(1, 70))
    {
        log_warn("Cannot read old configuration file (Ver%d.%2d)... renamed.\n",
                 file_major_version, file_minor_version);
        fclose(fp);
        rename_data(CONFIG_FILE);
        return 0;
    }
    //    fread(&config, sizeof(config_t), 1, fp);
    checked_fread((unsigned char *)&config, sizeof(config_t), fp,
                  NULL, CONFIG_FILE);
                                                  
    fclose(fp);
    return 1;
}


int save_config(void){
    FILE *fp;

    if ((fp = open_data(CONFIG_FILE, "wb", APPLICATION_ID, APPLICATION_ID_SIZE,
                        &major_version, &minor_version)) == NULL)
    {
        return 0;
    }
    fwrite(&config, sizeof(config_t), 1, fp);
    fclose(fp);
    return 1;
}


void draw_achievements(void){
    static int blink = 0;
    char buf[256];
    int i, kill_sum = 0, crush_sum = 0, shotdown_sum = 0; //, death_sum;
    enemytype_t et;

    for (et = et_start ; et < et_end ; et++){
        // GUNROID付属のBITSはそれぞれ個別カウントするが、
        // DRAGON_HEADはDRAGONの一部なので単体では数えない
//        if (et == et_dragon_head) continue;
        kill_sum += achievements.total_kill_type[et];
        crush_sum += achievements.total_crush_type[et];
        shotdown_sum += achievements.total_shotdown_type[et];
    }
    //    SetColor(5);
    set_color(ct_achievements);
    EzDrawStringB(320-12*12/2/2, 20, "ACHIEVEMENTS");
    for (i = 0 ; i < ACHIEVEMENTS_NUM ; i++){
        if (i-6 == et_dragon_head) continue;
        if (i < 6 || achievements.total_kill_type[i-6] > 0 ||
            achievements.total_crush_type[i-6] > 0 ||
            achievements.total_shotdown_type[i-6] > 0)
        {
            EzDrawStringB(60, ((i-6==et_core) ? i-1 : i)*20+40,
                          achievements_str[i]);
        }
        else {
            /*
            EzDrawStringB(60, ((i-6==et_core) ? i-1 : i)*20+40,
                          achievements_str[i]);
            */
            EzDrawStringB(60, ((i-6==et_core) ? i-1 : i)*20+40,
                          "  ???");
        }
        switch (i) {
        case 0:
            strcpy(buf, "ORIGINAL  EXTENDED      BOSS");
            break;
            //            sprintf(buf, "%8d", achievements.top_score_original);
        case 1:
            sprintf(buf, "%8d  %8d  %8d",
                    achievements.top_score[0],
                    achievements.top_score[1],
                    achievements.top_score[2]);
            //            sprintf(buf, "%8d", achievements.top_score_extended);
            break;
        case 2:
            sprintf(buf, "%8d  %8d  %8d",
                    achievements.top_stage[0],
                    achievements.top_stage[1],
                    achievements.top_stage[2]);
            //            sprintf(buf, "%8d", achievements.top_stage_original);
            break;
            //        case 3:
            //sprintf(buf, "%8d", achievements.top_stage_extended);
            // break;
        case 3:
            sprintf(buf, "%8d", achievements.total_play);
            break;
        case 4:
            strcpy(buf, "    KILL           CRUSH       SHOTDOWN");
            break;
        case 5:
            sprintf(buf, "%8d        %8d       %8d",
                    kill_sum, crush_sum, shotdown_sum);
            break;
        default:
            sprintf(buf, "%8d(%3d%%)  %8d(%3d%%) %8d(%3d%%)", 
                    achievements.total_kill_type[i-6],
                    (kill_sum == 0) ? 0 :
                    achievements.total_kill_type[i-6]*100 / kill_sum,
                    achievements.total_crush_type[i-6],
                    (crush_sum == 0) ? 0 :
                    achievements.total_crush_type[i-6]*100 / crush_sum,
                    achievements.total_shotdown_type[i-6],
                    (shotdown_sum == 0) ? 0 :
                    achievements.total_shotdown_type[i-6]*100 / shotdown_sum);
            break;
        }
        EzDrawStringB(240, ((i-6==et_core) ? i-1 : i)*20+40, buf);
    }
    if (grid_counter == 0){
        if (blink){
            blink = 0;
        }
        else {
            blink = 1;
        }
    }
    if (blink){
        EzDrawStringB(280, 464, "PUSH ANY KEY");
  }
}

void draw_config(void){
    int i, j, x;

    //    SetColor(4);
    set_color(ct_menu);
    EzDrawStringB(320-12*12/2/2, 20, "CONFIGURATION");
    for (i = 0 ; i < CONFIG_NUM ; i++){
        EzDrawStringB(160, i*32+60, config_str[i]);
    }
    for (i = 0 ; i < CONFIG_NUM+1 ; i++){
        for (j = 0 ; j < options_num[i] ; j++){
            EzDrawStringB(240+j*300/options_num[i], i*32+60,
                              options_str[i][j]);
            x = config.options[i] * 300 / options_num[i];
            //            EzDrawBoxB(240-4+x, i*32+60-20,
            //                       240-24+x+300/options_num[i], i*32+60+4);
            EzDrawLineB(240-4+x, i*32+60+4,
                       240-24+x+300/options_num[i], i*32+60+4);

        }
    }
}
 

void add_score(int point){
    AddUInt(score, point);
    update_top_score();
}


void update_top_score(void){
    // 通常のプレイならハイスコア更新
    if (!is_cheat){
        if (achievements.top_score[config.options[ci_stage]] < score){
            achievements.top_score[config.options[ci_stage]] = score;
        }
    }
}

void set_stage(int new_stage){
    stage = new_stage;
    printf("set stage %d %d %d\n", stage, midboss_index, boss_index);
    if (config.options[ci_stage] != gm_original){
        switch ((stage-1) % 3){
        case 0:
            group[midboss_index].type = et_bomber;
            group[boss_index].type = et_fortress;
            break;
        case 1:
            group[midboss_index].type = et_byte;
            group[boss_index].type = et_gunroid;
            break;
        case 2:
            group[midboss_index].type = et_wyvern;
            group[boss_index].type = et_dragon;
            break;
        default:
            printf("unkown boss in set_stage().\n");
            exit(1);
        }
    }
    update_top_stage();
}


void update_top_stage(void){
    // 通常のプレイならトップステージ更新
    if (!is_cheat){
        if (achievements.top_stage[config.options[ci_stage]] < (unsigned int)stage){
            achievements.top_stage[config.options[ci_stage]] = stage;
        }
    }
}


