#ifndef TDSHOOT_H
#define TDSHOOT_H
#include "config.h"
#include "3d.h"
#include "3dmodel.h"
#include "3ddraw.h"
#include "uty.h"

#define UINT_VALUE_MAX  999999999
#define AddUInt(v, a) (AddLimit(v, a, UINT_VALUE_MAX))
#define IncUInt(v) (IncLimit(v, UINT_VALUE_MAX))

#define GRIDSIZE 16

#define IMMORTAL_TIME 32
#define MYBULLET_MAX 16
#define ENEMYSHIP_MAX 16
#define ENEMYBULLET_MAX 12

#define MYSHIP_EXPLODE_MAX  8
#define ENEMYSHIP_EXPLODE_MAX  4

#define MYSHIP_SPEED          32
#define MYBULLET_SPEED        64
#define MYBULLET_SIZE         12

#define ENEMYBULLET_SIZE      8

#define SCOUT_XSPEED          32
#define SCOUT_ZSPEED          32
#define WALL_ZSPEED           GRIDSIZE
#define DRONE_XSPEED          16
#define DRONE_ZSPEED          32
#define DRONE_FIRE_RATE       32
#define DRONE_DIST            128
#define DRONE_BULLET_SPEED    16
#define FIGHTER_ZSPEED        48
#define FIGHTER_TURNOFFSET    160
#define FIGHTER_BULLET_SPEED  48
#define COPTER_XSPEED         (GRIDSIZE + 8)
#define COPTER_ZSPEED         (GRIDSIZE + 8)
#define COPTER_FIRE_RATE      24
#define COPTER_BULLET_SPEED   24
#define LANCER_ZSPEED         48
#define WALKER_XSPEED         16
#define WALKER_ZSPEED         GRIDSIZE
#define WALKER_FIRE_RATE      16
#define WALKER_BULLET_SPEED   32
#define TURRET_FIRE_ZPOS      ((ZMAX-ZMIN)*2/3 + ZMIN)
#define TURRET_BULLET_SPEED   32
#define TURRET_FIRE_RATE      16
#define FORTRESS_SPEED        4
#define FORTRESS_BULLET_SPEED 32
#define FORTRESS_FIRE_RATE    16
#define FORTRESS_BASELIFE 32
#define FORTRESS_ADDLIFE 16
#define BOMBER_BULLET_SPEED   32
#define BOMBER_FIRE_RATE      32
#define BOMBER_BASELIFE 32
#define BOMBER_ADDLIFE 16
#define WYVERN_BULLET_SPEED   24
#define WYVERN_SPEED 32
#define WYVERN_FIRE_RATE      16
#define WYVERN_BASELIFE 64
#define WYVERN_ADDLIFE 16
#define BYTE_NUM  9
#define BYTE_MOVESTEP 8
#define BYTE_SHIELD_ZPOS 500
#define BYTE_ASSAULT_ZPOS 500
#define BYTE_SHIELD_TIME 8
#define BYTE_BULLET_SPEED 32
#define BYTE_CHARGE_TIME 8
#define GUNROID_BASELIFE 48
#define GUNROID_ADDLIFE 16
#define GUNROID_DEFENCE_ZPOS   560
#define DRAGON_HEADLIFE 48
//#define DRAGON_HEADLIFE 2
#define DRAGON_BASELIFE DRAGON_HEADLIFE
// STAGEが進んだとき、頭のLIFE×頭の数だけの弾に耐え、
// 第2形態の方も強くなるから、
// 頭毎のLIFEの増加は控えめにしておく
#define DRAGON_ADDLIFE 8
#define DRAGON_HEAD_BASE_NUM 3
#define DRAGON_HEAD_MAX_NUM 8
#define DRAGON_HEAD_SPEED 16
#define DRAGON_HEAD_MOVE_RANGE 160
#define DRAGON_SPEED 4
#define DRAGON_ZPOS 560
#define BREATH_LEN 40
#define CORE_BASELIFE 80
#define CORE_ADDLIFE 16
#define CORE_SPEED 8
#define CORE_WANDER_COUNT 80
//#define CORE_WANDER_COUNT 8
#define CORE_FIRE_LASER_RATE 32
#define CORE_CHARGE_SPEED 16
#define CORE_BULLET_SPEED 24
#define LASER_MAX_LEN 4
#define LASER_MAX_NUM 8
#define LASER_ACCERATION 8


// GUNROIDのアサルトモードL, RはGUNROID体力の偶数奇数で切り替える
#define IsAssaultL()     (boss_life % 2 == 0)
// 同じボスの再出現回数(初回時は0)
#define BossRound()      (config.options[ci_stage] == gm_original ? \
                          (stage-1) : (stage-1)/3)

#define ShowHitBox()      (config.options[ci_hitbox])

typedef enum {
    gs_title, gs_playing, gs_demo, gs_achievements, gs_config
} gamestate_t;


typedef enum {
  ss_alive, ss_exploding, ss_dead,
} shipstate_t;

typedef enum {
  bs_alive, bs_exploding, bs_dead
} bulletstate_t;

typedef struct {
  point3d_t pos;
  shipstate_t state;
  int xa, za;
  int explode_motion;
  int left;
} myship_t;

typedef enum {
    ea_init = 0,

    ea_dragon_head_set_dest = 0,
    ea_dragon_head_move,
    ea_dragon_head_open_mouth,
    ea_dragon_head_breath,
    ea_dragon_head_close_mouth,

    ea_dragon_engage = 0,
    ea_dragon_wander,

    ea_core_takeoff = 0,
    ea_core_yaw_left,
    ea_core_yaw_right,
    ea_core_setup_charge,
    ea_core_charge,
    ea_core_retreat,
    ea_core_open_mouth,
    ea_core_laser,
    ea_core_close_mouth,
} enemyship_action_t;

typedef struct {
    point3d_t pos;
    shipstate_t state;
    int explode_motion;
    enemytype_t type;
    //  dir_t dir;
    int xspeed, zspeed;
    int motion;          // モーションパターン
    int fire_counter;
    int fire_rate;
    int counter;   // 汎用カウンタ
    enemyship_action_t action; // move_xxxでの状態遷移用
    int can_hit; // 当たり判定ありなら非0
    int group_no;
} enemyship_t;

typedef struct {
  char *name;
  int score;
} enemydata_t;

typedef struct {
    point3d_t pos;
    int xa, za;
    bulletstate_t state;
    enemytype_t owner_type;
} bullet_t;

typedef struct {
    point3d_t t, l, r, b;
} breath_t;

typedef struct {
    point3d_t t;
} laser_node_t;

typedef struct {
    int is_valid;
    laser_node_t nodes[LASER_MAX_LEN];
    int head, tail;
    int vx, vz;
} laser_t;

#define SuccLaserIx(ix) (((ix)+1) % LASER_MAX_LEN)
#define IncLaserIx(ix) (ix = SuccLaserIx(ix))


typedef struct {
  enemytype_t type;
  int enemy_num;  // typeの敵が何機出るか
  int enemy_max;  // 画面内に何機まで出るか
} group_t;

typedef enum {
    cm_start, cm_achievements, cm_config, cm_quit
} cmd_t;

typedef enum {
    ci_stage, ci_myship, ci_autofire, ci_hitbox, ci_colorset, ci_view,
    ci_end
} config_item_t;

typedef struct {
    unsigned char options[ci_end+1];
} config_t; 

typedef enum{
    gm_original, gm_extended, gm_boss, gm_reserve = 7
} game_mode_t;

typedef struct {
    unsigned int top_score[gm_reserve];
    unsigned int top_stage[gm_reserve];
    unsigned int total_play;
    unsigned int total_kill_type[et_reserve];
    unsigned int total_crush_type[et_reserve]; // 各敵に接触して死んだ回数
    unsigned int total_shotdown_type[et_reserve]; // 各敵の弾で死んだ回数
} achievements_t;

extern breath_t breath[5][BREATH_LEN];
extern config_t config;

void title_keyh(int key);
void title_timerh(void);
void achievements_keyh(int key);
void achievements_timerh(void);
void config_keyh(int key);
void config_timerh(void);
void game_keyh(int key);
void game_timerh(void);

void init(int argc, char *argv[]);
void init_title(void);
void init_game(void);
void init_achievements(void);
void init_config(void);

void init_myship(void);
void init_boss_life(int base_life, int add_life);

int is_hit(point3d_t p1, point3d_t p2, int xrange, int zrange);

void set_view(view_mode_t new_view_mode);
void toggle_view_mode(void);

void draw(void);
void draw_ground(void);
void animate_sensor(void);
void draw_sensor(void);
void draw_info(void);

void draw_ship_shadows(void);
void draw_myship(void);
void draw_mybullet(bullet_t *bullet);
void draw_enemyship(enemyship_t *ship);
void draw_enemybullet(bullet_t *bullet);
void draw_neck(enemyship_t *head);
void draw_breath(enemyship_t *head);
void check_hit_breath(enemyship_t *head);
void draw_lasers(void);
void draw_laser(laser_t *laser);
void create_enemyship(void);
int allocate_enemyship(void);
int find_alive_enemyship(enemytype_t enemytype);
void delete_enemyship(enemyship_t *ship);
void create_scout(enemyship_t *ship);
void create_wall(enemyship_t *ship);
void create_drone(enemyship_t *ship);
void create_fighter(enemyship_t *ship);
void create_copter(enemyship_t *ship);
void create_lancer(enemyship_t *ship);
void create_walker(enemyship_t *ship);
void create_turret(enemyship_t *ship);
void create_fortress(enemyship_t *ship);
void create_bomber(enemyship_t *ship);
void create_wyvern(enemyship_t *ship);
void create_byte(enemyship_t *ship);
void create_gunroid(enemyship_t *ship);
void create_dragon_head(enemyship_t *ship);
void create_dragon(enemyship_t *ship);
void create_core(enemyship_t *ship);

void check_hit_lasers(void);
void check_hit_laser(laser_t *laser);
void move_lasers(void);
void move_laser(laser_t *laser);
void move_myship(void);
void move_mybullet(bullet_t *bullet);
void move_enemyship(enemyship_t *ship, int no);
void move_scout(enemyship_t *ship);
void move_wall(enemyship_t *ship);
void move_drone(enemyship_t *ship);
void move_fighter(enemyship_t *ship);
void move_copter(enemyship_t *ship);
void move_lancer(enemyship_t *ship);
void move_walker(enemyship_t *ship);
void move_turret(enemyship_t *ship);
void move_fortress(enemyship_t *ship);
void move_bomber(enemyship_t *ship);
void move_wyvern(enemyship_t *ship);
void move_byte(enemyship_t *ship, int no);
void move_gunroid(enemyship_t *ship);
void move_dragon_head(enemyship_t *ship);
void move_dragon(enemyship_t *ship);
void move_core(enemyship_t *ship);
void motion_rattle_core(enemyship_t *ship);

void move_enemybullet(bullet_t *bullet);

void check_hit_enemyship(enemyship_t *ship);
void check_hit_enemybullet(bullet_t *bullet);


void enemy_fire_forward(point3d_t pos, int speed, enemytype_t type);
void enemy_fire_aiming(point3d_t pos, int speed, enemytype_t type);
void enemy_fire_aiming(point3d_t pos, int speed, enemytype_t type);
void enemy_fire_spread(point3d_t pos, int speed, int num, enemytype_t type);
int enemy_create_bullet(point3d_t pos, enemytype_t type);
void create_lasers(enemyship_t *enemy);
void create_laser(enemyship_t *enemy, int vx0, int vz0);


void draw_achievements(void);
void draw_config(void);
void fire(void);
void quit(void);
int load_achievements(void);
int save_achievements(void);
int load_config(void);
int save_config(void);
void add_score(int point);
void update_top_score(void);
void set_stage(int new_stage);
void update_top_stage(void);
void redraw(void);

#endif
